from __future__ import annotations

import math
import random
from copy import deepcopy
from typing import Union

import numpy as np
from holisticai.bias.mitigation.commons.disparate_impact_remover._categorical_feature import CategoricalFeature
from holisticai.bias.mitigation.commons.disparate_impact_remover._sparse_list import SparseList


def get_categories_count_norm(categories, all_stratified_groups, count_dict, group_features):
    """
    Get the normalized count for each category , where normalized count is count divided by
    the number of people in that group.


    Parameters
    ----------
    categories : list
        The list of categories.
    all_stratified_groups : list
        The list of stratified groups.
    count_dict : dict
        The dictionary containing the count of observations in each category.
    group_features : dict
        The dictionary containing the CategoricalFeature objects for each group.

    Returns
    -------
    dict
        The dictionary containing the normalized count for each category.
    """
    norm = {
        cat: SparseList(
            data=(
                count_dict[cat][i] * (1.0 / len(group_features[group].data)) if group_features[group].data else 0.0
                for i, group in enumerate(all_stratified_groups)
            )
        )
        for cat in categories
    }
    return norm


def gen_desired_dist(group_index, cat, col_id, median, repair_level, norm_counts, feature_to_remove, mode):
    """
    Generate the desired distribution for each category in a group.

    Parameters
    ----------
    group_index : int
        The index of the group.
    cat : str
        The category.
    col_id : int
        The column index.
    median : dict
        The dictionary containing the median value for each category.
    repair_level : float
        The repair level.
    norm_counts : dict
        The dictionary containing the normalized count for each category.
    feature_to_remove : int
        The feature to remove.
    mode : str
        The mode value.

    Returns
    -------
    float
        The desired distribution for the category.
    """
    if feature_to_remove == col_id:
        return 1 if cat == mode else (1 - repair_level) * norm_counts[cat][group_index]
    return (1 - repair_level) * norm_counts[cat][group_index] + (repair_level * median[cat])


def assign_overflow(all_stratified_groups, categories, overflow, group_features, repair_generator):
    """
    Assign the overflow to the desired categories based on the group's desired distribution.

    Parameters
    ----------
    all_stratified_groups : list
        The list of stratified groups.
    categories : list
        The list of categories.
    overflow : dict
        The dictionary containing the overflow for each group.
    group_features : dict
        The dictionary containing the CategoricalFeature objects for each group.
    repair_generator : function
        The repair generator function.

    Returns
    -------
    dict
        The dictionary containing the assigned overflow for each group.
    dict
        The dictionary containing the desired distribution for each category.
    dict
        The dictionary containing the feature after overflow assignment.
    """
    feature = deepcopy(group_features)
    assigned_overflow = {}
    desired_dict_list = {}
    for group_index, group in enumerate(all_stratified_groups):
        # Calculate the category proportions.
        def dist_generator(cat):
            return repair_generator(group_index, cat)

        cat_props = list(map(dist_generator, categories))

        if all(elem == 0 for elem in cat_props):  # TODO: Check that this is correct!
            cat_props = [1.0 / len(cat_props)] * len(cat_props)
        s = float(sum(cat_props))
        cat_props = [elem / s for elem in cat_props]
        desired_dict_list[group] = cat_props
        assigned_overflow[group] = {}
        for i in range(int(overflow[group])):
            distribution_list = desired_dict_list[group]
            number = random.uniform(0, 1)
            cat_index = 0
            tally = 0
            for j in range(len(distribution_list)):
                value = distribution_list[j]
                if number < (tally + value):
                    cat_index = j
                    break
                tally += value
            assigned_overflow[group][i] = categories[cat_index]
        # Actually do the assignment
        count = 0
        for i, value in enumerate(group_features[group].data):
            if value == 0:
                (feature[group].data)[i] = assigned_overflow[group][count]
                count += 1
    return feature, assigned_overflow, desired_dict_list


def get_categories_count(categories, all_stratified_groups, group_feature):
    """
    Get the count for each category.

    Parameters
    ----------
    categories : list
        The list of categories.
    all_stratified_groups : list
        The list of stratified groups.
    group_feature : dict
        The dictionary containing the CategoricalFeature objects for each group.

    Returns
    -------
    dict
        The dictionary containing the count for each category.
    """
    count_dict = {
        cat: SparseList(
            data=(
                group_feature[group].category_count[cat] if cat in group_feature[group].category_count else 0
                for group in all_stratified_groups
            )
        )
        for cat in categories
    }

    return count_dict


def gen_desired_count(group_index, group, category, median, group_features, repair_level, categories_count):
    """
    Generate the desired count for each category in a group.

    Parameters
    ----------
    group_index : int
        The index of the group.
    group : str
        The group.
    category : str
        The category.
    median : dict
        The dictionary containing the median value for each category.
    group_features : dict
        The dictionary containing the CategoricalFeature objects for each group.
    repair_level : float
        The repair level.
    categories_count : dict
        The dictionary containing the count for each category.

    Returns
    -------
    int
        The desired count for the category.
    """
    med = median[category]
    size = len(group_features[group].data)
    count = categories_count[category][group_index]
    des_count = math.floor(((1 - repair_level) * count) + (repair_level) * med * size)
    return des_count


def flow_on_group_features(all_stratified_groups, group_features, repair_generator):
    """
    Flow on group features. Run Max-flow to distribute as many observations to categories as possible.
    Overflow are those observations that are left over after max-flow.

    Parameters
    ----------
    all_stratified_groups : list
        The list of stratified groups.
    group_features : dict
        The dictionary containing the CategoricalFeature objects for each group.
    repair_generator : function
        The repair generator function.

    Returns
    -------
    dict
        The dictionary containing the new feature after repair.
    dict
        The dictionary containing the overflow for each group.Description
    """
    dict1 = {}
    dict2 = {}
    for i, group in enumerate(all_stratified_groups):
        feature = group_features[group]

        def count_generator(category):
            return repair_generator(i, group, category)

        DG = feature.create_graph(count_generator)
        new_feature, overflow = feature.repair(DG)
        dict2[group] = overflow
        dict1[group] = new_feature

    return dict1, dict2


def get_count(cat, group_feature_category_count):
    """
    Get the count for a category.

    Parameters
    ----------
    cat : str
        The category.
    group_feature_category_count : dict
        The dictionary containing the count for each category.

    Returns
    -------
    int
        The count for the category.
    """
    if cat in group_feature_category_count:
        return group_feature_category_count[cat]
    return 0


def get_count_norm(count, group_feature_data):
    """
    Get the normalized count for a category.

    Parameters
    ----------
    count : int
        The count for the category.
    group_feature_data : list
        The list of data for the category.

    Returns
    -------
    float
        The normalized count for the category.
    """
    if group_feature_data:
        return count * (1.0 / len(group_feature_data))
    return 0.0


def get_column_type(values: Union[list, np.ndarray]):
    """
    Get the type of the column.

    Parameters
    ----------
    values : list or np.ndarray
        The list of values.

    Returns
    -------
    type
        The type of the column.
    """
    if all(isinstance(value, float) for value in values):
        return float

    if all(isinstance(value, int) for value in values):
        return int
    return str


def get_median(values, kdd):
    """
    Get the median of a list of values.

    Parameters
    ----------
    values : list
        The list of values.
    kdd : bool
        Whether to use the KDD method.

    Returns
    -------
    float
        The median of the list of values.
    """
    if not values:
        raise ValueError("Cannot calculate median of list with no values!")

    sorted_values = deepcopy(values)
    sorted_values.sort()  # Not calling `sorted` b/c `sorted_values` may not be list.

    if kdd:
        return sorted_values[len(values) // 2]

    if len(values) % 2 == 0:
        return sorted_values[len(values) // 2 - 1]

    return sorted_values[len(values) // 2]


def get_group_data(all_stratified_groups, stratified_group_data, col_id):
    """
    Get the group data for a column.

    Parameters
    ----------
    all_stratified_groups : list
        The list of stratified groups.
    stratified_group_data : dict
        The dictionary containing the stratified group data.
    col_id : int
        The column index.

    Returns
    -------
    dict
        The dictionary containing the group data for the column.
    """
    group_features = {}
    for group in all_stratified_groups:
        points = [(i, val) for indices, val in stratified_group_data[group][col_id] for i in indices]
        points = sorted(points, key=lambda x: x[0])  # Sort by index
        values = [value for _, value in points]

        # send values to CategoricalFeature object, which bins the data into categories
        group_features[group] = CategoricalFeature(values)
    return group_features


# Find the median normalized count for each category
def get_median_per_category(categories, categories_count_norm):
    """
    Get the median normalized count for each category.

    Parameters
    ----------
    categories : list
        The list of categories.
    categories_count_norm : dict
        The dictionary containing the normalized count for each category.

    Returns
    -------
    dict
        The dictionary containing the median normalized count for each category.
    """
    return {cat: get_median(categories_count_norm[cat], False) for cat in categories}


def get_mode(values):
    """
    Get the mode of a list of values.

    Parameters
    ----------
    values : list
        The list of values.

    Returns
    -------
    any
        The mode of the list of values.
    """
    counts = {}
    for value in values:
        counts[value] = 1 if value not in counts else counts[value] + 1
    mode_tuple = max(list(counts.items()), key=lambda tup: tup[1])
    return mode_tuple[0]


def make_histogram_bins(bin_size_calculator, data, col_id):
    """
    Make histogram bins for a column.

    Parameters
    ----------
    bin_size_calculator : function
        The bin size calculator function.
    data : list
        The input dataset.
    col_id : int
        The column index.

    Returns
    -------
    list
        The list of histogram bins.
    """
    feature_vals = [row[col_id] for row in data]
    bin_range = bin_size_calculator(feature_vals)

    if bin_range == 0.0:
        bin_range = 1.0

    data_tuples = list(enumerate(data))  # [(0,row), (1,row'), (2,row''), ... ]
    sorted_data_tuples = sorted(data_tuples, key=lambda tup: tup[1][col_id])

    max_val = max(data, key=lambda datum: datum[col_id])[col_id]
    min_val = min(data, key=lambda datum: datum[col_id])[col_id]

    index_bins = []
    val_ranges = []
    curr = min_val
    while curr <= max_val:
        index_bins.append([])
        val_ranges.append((curr, curr + bin_range))
        curr += bin_range

    for row_index, row in sorted_data_tuples:
        for bin_num, val_range in enumerate(val_ranges):
            if val_range[0] <= row[col_id] < val_range[1]:
                index_bins[bin_num].append(row_index)
                break

    index_bins = [b for b in index_bins if b]

    return index_bins


def freedman_diaconis_bin_size(feature_values):
    """
    Calculate the bin size using the Freedman-Diaconis rule.

    Parameters
    ----------
    feature_values : list
        The list of feature values.

    Returns
    -------
    float
        The bin size.
    """
    q75, q25 = np.percentile(feature_values, [75, 25])
    IQR = q75 - q25
    return 2.0 * IQR * len(feature_values) ** (-1.0 / 3.0)



from holisticai.efficacy.metrics._classification import classification_efficacy_metrics, confusion_matrix
from holisticai.efficacy.metrics._clustering import clustering_efficacy_metrics
from holisticai.efficacy.metrics._multiclass import multiclassification_efficacy_metrics
from holisticai.efficacy.metrics._regression import regression_efficacy_metrics, rmse_score, smape

__all__ = [
    "classification_efficacy_metrics",
    "confusion_matrix",
    "clustering_efficacy_metrics",
    "regression_efficacy_metrics",
    "multiclassification_efficacy_metrics",
    "rmse_score",
    "smape",
]



# Imports
import numpy as np
import pandas as pd
from numpy.testing import assert_approx_equal

# Classification
from holisticai.bias.metrics import (
    abroca,
    accuracy_diff,
    average_odds_diff,
    cohen_d,
    disparate_impact,
    equal_opportunity_diff,
    false_negative_rate_diff,
    false_positive_rate_diff,
    four_fifths,
    statistical_parity,
    success_rate,
    true_negative_rate_diff,
    z_test_diff,
    z_test_ratio,
)

# Formatting
from holisticai.utils import extract_columns

# Dataset
from tests.bias.utils import load_bias_classification_data

df_c = load_bias_classification_data()

# Format data into numpy
group_a, group_b, y_pred, y_true = extract_columns(
    df_c, cols=["group_a", "group_b", "y_pred", "y_true"]
)


def test_statistical_parity():
    """test statistical_parity"""
    a = statistical_parity(group_a, group_b, y_pred)
    b = 5 / 12
    assert_approx_equal(a, b)


def test_disparate_impact():
    """test disparate_impact"""
    a = disparate_impact(group_a, group_b, y_pred)
    b = 9 / 4
    assert_approx_equal(a, b)


def test_success_rate():
    sr_a = (
        df_c[df_c["group_a"] == 1]["y_pred"].sum() / df_c[df_c["group_a"] == 1].shape[0]
    )
    assert success_rate(group_a, group_b, y_pred)["sr_a"] == sr_a


def test_four_fifths():
    """test four_fifths"""
    a = four_fifths(group_a, group_b, y_pred)
    b = 4 / 9
    assert_approx_equal(a, b)


def test_cohen_d():
    """test cohen_d"""
    a = cohen_d(group_a, group_b, y_pred)
    b = 0.9109750373485539
    assert_approx_equal(a, b)


def test_z_test_diff():
    """test z_test_diff"""
    a = z_test_diff(group_a, group_b, y_pred)
    b = 1.290994449
    assert_approx_equal(a, b)


def test_z_test_ratio():
    """test z_test_ratio"""
    a = z_test_ratio(group_a, group_b, y_pred)
    b = 1.256287689
    assert_approx_equal(a, b)


def test_equal_opportunity_diff():
    """test equal_oppotunity_diff"""
    a = equal_opportunity_diff(group_a, group_b, y_pred, y_true)
    b = 1 / 3
    assert_approx_equal(a, b)


def false_negative_rate_diff():
    """test false_negative_rate_diff"""
    a = false_negative_rate_diff(group_a, group_b, y_pred, y_true)
    b = -1 / 3
    assert_approx_equal(a, b)


def true_negative_rate_diff():
    """test true_negative_rate_diff"""
    a = true_negative_rate_diff(group_a, group_b, y_pred, y_true)
    b = -1 / 2
    assert_approx_equal(a, b)


def test_false_positive_rate_diff():
    """test false_positive_rate_diff"""
    a = false_positive_rate_diff(group_a, group_b, y_pred, y_true)
    b = 1 / 2
    assert_approx_equal(a, b)


def test_average_odds_diff():
    """test average_odds_diff"""
    a = average_odds_diff(group_a, group_b, y_pred, y_true)
    b = 5 / 12
    assert_approx_equal(a, b)


def test_accuracy_diff():
    """test accuracy_diff"""
    a = accuracy_diff(group_a, group_b, y_pred, y_true)
    b = 3 / 4 - 5 / 6
    assert_approx_equal(a, b)



import copy
from typing import Any

import pandas as pd
from holisticai.bias.mitigation.inprocessing.grid_search._grid_generator import GridGenerator
from joblib import Parallel, delayed


class GridSearchAlgorithm:
    """
    Estimator to perform a grid search given a blackbox estimator algorithm.
    The approach used is taken from section 3.4 of "A reductions approach to fair classification"
    Enabled for more than two sensitive feature values

    Reference
    ---------
    Agarwal, Alekh, et al. "A reductions approach to fair classification."
    International Conference on Machine Learning. PMLR, 2018.
    """

    def __init__(
        self,
        estimator: Any,
        constraint: Any,
        constraint_weight: float = 0.5,
        grid_size: int = 20,
        grid_limit: int = 2,
        n_jobs=-1,
        verbose=0,
    ):
        """
        Init GridSearchAlgorithm object

        Parameters
        ----------
        estimator : object
            An estimator implementing methods :code:`fit(X, y, sample_weight=sample_weight)` and
            :code:`predict(X)`, where `X` is the matrix of features, `y` is the
            vector of labels, and `sample_weight` is a vector of weights.
            In binary classification labels `y` and predictions returned by
            :code:`predict(X)` are either 0 or 1.

        constraint : ClassificationConstraint
            The disparity constraint.

        constraint_weight : float
            Specifies the relative weight put on the constraint violation when selecting the
            best model. The weight placed on the error rate will be :code:`1-constraint_weight`.

        grid_size: int
            Number of columns to be generated in the grid.

        grid_limit : float
            Range of the values in the grid generated.

        verbose : int
            If >0, will show progress percentage.
        """
        self.estimator: Any = estimator
        self.constraint: Any = constraint
        self.objective: Any = constraint.default_objective()
        self.grid_limit: int = grid_limit
        self.grid_size: int = grid_size
        self.constraint_weight = constraint_weight
        self.n_jobs = n_jobs
        self.verbose = verbose

    def fit(self, X: Any, y: Any, sensitive_features: Any):
        """
        Fit model using Grid Search Algorithm.

        Parameters
        ----------
        X : matrix-like
            Input matrix.

        y : numpy array
            Target vector.

        sensitive_features : numpy array
            Matrix where each column is a sensitive feature e.g. [col_1=group_a, col_2=group_b].

        Returns
        -------
        GridSearchAlgorithm
            The same object.
        """
        self._load_data(X, y, sensitive_features)
        grid = self._generate_grid()

        results = Parallel(n_jobs=self.n_jobs, verbose=self.verbose)(
            delayed(self.evaluate_candidate)(
                X,
                grid[col_name],
            )
            for col_name in grid
        )

        results = pd.DataFrame(results)
        results["col_name"] = grid.columns
        best_idx = results["loss"].idxmin()
        best_colname = results["col_name"].iloc[best_idx]
        self.bet_predictor = self._fit_estimator(X, grid[best_colname])
        return self

    def evaluate_candidate(self, X, lambda_vec):
        current_estimator = self._fit_estimator(X, lambda_vec)

        def predict_fn(X: Any) -> Any:
            return current_estimator.predict(X)

        objective = self.objective.gamma(predict_fn).iloc[0]
        gamma = self.constraint.gamma(predict_fn)
        loss = (1 - self.constraint_weight) * objective + self.constraint_weight * gamma.max()
        return {"loss": loss, "lambda_vec": lambda_vec}

    def _load_data(self, X: Any, y: Any, sensitive_features: Any):
        self.constraint.load_data(X, y, sensitive_features)
        self.objective.load_data(X, y, sensitive_features)

    def _generate_grid(self) -> Any:
        neg_allowed = self.constraint.neg_basis_present
        objective_in_the_span = self.constraint.default_objective_lambda_vec is not None

        generator = GridGenerator(
            grid_size=self.grid_size,
            grid_limit=self.grid_limit,
            neg_allowed=neg_allowed,
            force_L1_norm=objective_in_the_span,
        )

        return generator.generate_grid(self.constraint)

    def _fit_estimator(self, X: Any, lambda_vec: Any):
        weights = self._compute_weights(lambda_vec)
        y_reduction, weights = self._get_reduction(weights)
        current_estimator = copy.deepcopy(self.estimator)
        current_estimator.fit(X, y_reduction, sample_weight=weights)
        return current_estimator

    def _compute_weights(self, lambda_vec: Any) -> Any:
        weights = self.constraint.signed_weights(lambda_vec)
        if not self.constraint.default_objective_lambda_vec:
            weights += self.objective.signed_weights()
        return weights

    def _get_reduction(self, weights: Any) -> tuple:
        if self.constraint.PROBLEM_TYPE == "classification":
            y_reduction = 1 * (weights > 0)
            weights = weights.abs()
        else:
            y_reduction = self.constraint.y_as_series
        return y_reduction, weights

    def predict(self, X: Any) -> Any:
        return self.bet_predictor.predict(X)

    def predict_proba(self, X: Any) -> Any:
        return self.bet_predictor.predict_proba(X)



# Imports
import numpy as np
from numpy.testing import assert_approx_equal, assert_array_almost_equal

# Multiclass
from holisticai.bias.metrics import (
    accuracy_matrix,
    confusion_tensor,
    frequency_matrix,
    multiclass_average_odds,
    multiclass_equality_of_opp,
    multiclass_statistical_parity,
    multiclass_true_rates,
    precision_matrix,
    recall_matrix,
)

# Format data into numpy for each task
g_min = np.array([1, 1, 1, 1, 0, 0, 0, 0, 0, 0])
g_maj = np.array([0, 0, 0, 0, 1, 1, 1, 1, 1, 1])

p = np.array(["A", "A", "A", "A", "B", "B", "B", "B", "C", "C"])
y_pred_mc = np.array([0, 1, 2, 0, 1, 2, 0, 1, 2, 0])
y_true_mc = np.array([0, 1, 1, 0, 1, 0, 2, 1, 2, 1])

p2 = np.array(["A", "A", "A", "A", "B", "B", "B", "B", "C", "C", "C"])
y_pred2 = np.array([0, 2, 1, 0, 2, 1, 0, 1, 1, 2, 0])


p_bin = np.array([0, 0, 0, 0, 1, 1, 1, 1, 1, 1])
y_pred = np.array([0, 1, 1, 0, 1, 1, 0, 1, 1, 0])
y_true = np.array([0, 1, 1, 0, 1, 0, 1, 1, 1, 1])


def test_confusion_tensor():
    """test confusion_tensor"""
    a = confusion_tensor(p, y_pred_mc, y_true_mc, as_tensor=True)
    b = np.array(
        [
            [[2.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 1.0, 0.0]],
            [[0.0, 0.0, 1.0], [0.0, 2.0, 0.0], [1.0, 0.0, 0.0]],
            [[0.0, 1.0, 0.0], [0.0, 0.0, 0.0], [0.0, 0.0, 1.0]],
        ]
    )
    assert_array_almost_equal(a, b)


def test_frequency_matrix():
    """test frequency_matrix"""
    a = frequency_matrix(p, y_pred_mc).to_numpy()
    b = np.array([[0.5, 0.25, 0.25], [0.25, 0.5, 0.25], [0.5, 0.0, 0.5]])
    assert_array_almost_equal(a, b)


def test_accuracy_matrix():
    """test accuracy_matrix"""
    a = accuracy_matrix(p, y_pred_mc, y_true_mc).to_numpy()
    b = np.array([[1.0, 0.5, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, 1.0]])
    assert_array_almost_equal(a, b)


def test_precision_matrix():
    """test precision_matrix"""
    a = precision_matrix(p, y_pred_mc, y_true_mc).to_numpy()
    b = np.array([[1.0, 0.5, np.nan], [0.0, 1.0, 0.0], [np.nan, 0.0, 1.0]])
    assert_array_almost_equal(a, b)


def test_recall_matrix():
    """test recall_matrix"""
    a = recall_matrix(p, y_pred_mc, y_true_mc).to_numpy()
    b = np.array([[1.0, 1.0, 0.0], [0.0, 1.0, 0.0], [0.0, np.nan, 1.0]])
    assert_array_almost_equal(a, b)


def test_multiclass_equality_of_opp():
    """test multiclass equality of opportunity"""
    a = multiclass_equality_of_opp(p_bin, y_pred, y_true)
    b = 0.7
    assert_array_almost_equal(a, b)

    a = multiclass_equality_of_opp(p, y_pred_mc, y_true_mc)
    b = 0.5308628148148148 * 1.5
    assert_array_almost_equal(a, b, 5)


def test_multiclass_average_odds():
    """test multiclass average odds"""
    a = multiclass_average_odds(p_bin, y_pred, y_true)
    b = 0.3
    assert_array_almost_equal(a, b)

    a = multiclass_average_odds(p, y_pred_mc, y_true_mc)
    b = (1 / 9) * 1.5
    assert_array_almost_equal(a, b)

    a = multiclass_average_odds(p2, y_pred2, y_pred2)
    b = 0
    assert_array_almost_equal(a, b)


def test_multiclass_true_rates():
    """test multiclass true rates"""
    a = multiclass_true_rates(p_bin, y_pred, y_true)
    b = 0.7
    assert_array_almost_equal(a, b)

    a = multiclass_true_rates(p, y_pred_mc, y_true_mc)
    b = 2 / 3
    assert_array_almost_equal(a, b)

    a = multiclass_true_rates(p2, y_pred2, y_pred2)
    b = 0
    assert_array_almost_equal(a, b)


def test_multiclass_statistical_parity():
    """test multiclass statistical parity"""
    a = multiclass_statistical_parity(p_bin, y_pred)
    b = 1 / 6
    assert_array_almost_equal(a, b)

    a = multiclass_statistical_parity(p, y_pred_mc)
    b = 2 * 3 / 18
    assert_array_almost_equal(a, b)



from holisticai.explainability.metrics.global_importance._alpha_score import (
    AlphaScore,
    alpha_score,
)
from holisticai.explainability.metrics.global_importance._importance_spread import (
    SpreadDivergence,
    SpreadRatio,
    spread_divergence,
    spread_ratio,
)
from holisticai.explainability.metrics.global_importance._position_parity import PositionParity, position_parity
from holisticai.explainability.metrics.global_importance._rank_alignment import RankAlignment, rank_alignment
from holisticai.explainability.metrics.global_importance._surrogate import (
    surrogate_accuracy_score,
    surrogate_mean_squared_error,
)
from holisticai.explainability.metrics.global_importance._xai_ease_score import XAIEaseScore, xai_ease_score

__all__ = [
    "AlphaScore",
    "XAIEaseScore",
    "PositionParity",
    "RankAlignment",
    "SpreadRatio",
    "SpreadDivergence",
    "alpha_score",
    "xai_ease_score",
    "position_parity",
    "rank_alignment",
    "spread_ratio",
    "spread_divergence",
    "surrogate_accuracy_score",
    "surrogate_mean_squared_error",
]



from __future__ import annotations

from typing import TYPE_CHECKING, Literal, Union

import pandas as pd
from sklearn.model_selection import train_test_split

from holisticai.utils.obj_rep.object_repr import generate_html_for_generic_object

if TYPE_CHECKING:
    from collections.abc import Iterable

import json

import numpy as np
from numpy.random import RandomState


class DatasetDict(dict):
    """
    A dictionary-like class that represents a collection of datasets. Usually, the keys are train, validation, and test.

    Parameters:
    -----------
    datasets : dict
        A dictionary containing the datasets, where the keys are the names of the datasets and the values are the datasets themselves.

    Methods:
    --------

    __getitem__(key)
        Returns a Dataset object corresponding to the given key.
    """

    def __init__(self, **datasets):
        self.datasets = datasets

    def __getitem__(self, key):
        return self.datasets[key]

    def __repr__(self):
        datasets_repr = ",\n    ".join(f"{name}: {dataset}" for name, dataset in self.datasets.items())
        return f"DatasetDict({{\n    {datasets_repr}\n}})"

    def _repr_html_(self):
        nested_objs = []
        for name, dataset in self.datasets.items():
            repr_info = dataset.repr_info()
            repr_info["name"] = name
            nested_objs.append(repr_info)
        # Example usage
        obj = {"dtype": "DatasetDict", "attributes": {}, "nested_objects": nested_objs}
        return generate_html_for_generic_object(obj, feature_columns=5)


class GroupByDataset:
    """
    A class representing a Grouped Dataset.

    Parameters:
    -----------
    groupby_obj : pandas.core.groupby.GroupBy
        The pandas GroupBy object representing the grouped dataset.

    Attributes:
    -----------
    grouped_names : list
        A list of the names of the groups in the dataset.
    features : list
        A list of the unique features in the dataset.
    ngroups : int
        The number of groups in the dataset.
    random_state : numpy.random.RandomState
        The random state object used for sampling.

    """

    def __init__(self, groupby_obj):
        self.groupby_obj = groupby_obj
        self.grouped_names = [feature for feature, _ in self.groupby_obj.keys]
        self.features = self.groupby_obj.obj.columns.get_level_values("features").unique().tolist()
        self.ngroups = self.groupby_obj.ngroups
        self.random_state = np.random.RandomState()

    def head(self, k):
        """Returns the first k rows of each group in the dataset."""
        return Dataset(self.groupby_obj.head(k))

    def sample(self, n, random_state=None):
        """Returns a random sample of n rows from each group in the dataset."""
        if random_state is None:
            random_state = self.random_state
        return Dataset(
            self.groupby_obj.apply(lambda x: sample_n(x, n, random_state=random_state)).reset_index(drop=True)
        )

    def __iter__(self):
        """Iterates over the groups in the dataset."""
        for group_name, groupby_obj_batch in self.groupby_obj:
            yield group_name, Dataset(groupby_obj_batch)

    def count(self):
        dss = []
        for group_name, groupby_obj_batch in self.groupby_obj:
            data = dict(zip(self.grouped_names, group_name))
            data["group_size"] = len(groupby_obj_batch)
            dss.append(data)
        return pd.DataFrame(dss)

    def __repr__(self):
        """Returns a string representation of the GroupByDataset."""
        return (
            f"GroupByDataset({{\n"
            f"        names: {self.grouped_names},\n"
            f"        features: {self.features},\n"
            f"        count: {self.ngroups}\n"
            f"    }})"
        )

    def _repr_html_(self):
        """Returns an HTML representation of the GroupByDataset."""
        obj = {
            "dtype": "GroupByDataset",
            "attributes": {
                "count": self.ngroups,
                "grouped_names": self.grouped_names,
                "Features": [" , ".join(self.features)],
            },
        }
        return generate_html_for_generic_object(obj, feature_columns=5)


def dataframe_to_level_dict_with_series(df, row_index):
    """
    Converts a DataFrame with two-level columns into a dictionary where:
    - Level 0 names become dictionary keys.
    - Corresponding DataFrames (without level 0) are values.
    - Level 1 names become new column names.

    Args:
        df (pandas.DataFrame): The DataFrame to convert.
        row_index (int): The index of the row to use as the Series.

    Returns:
        dict: The resulting dictionary with level 0 keys and DataFrames as values.
    """
    num_levels = 2
    if not isinstance(df.columns, pd.MultiIndex) or len(df.columns.levels) != num_levels:
        msg = "DataFrame must have MultiIndex columns with two levels."
        raise ValueError(msg)

    data = {}
    for level_0_name in df.columns.levels[0]:
        feature = df[level_0_name]
        if feature.shape[1] == 1:
            data[level_0_name] = feature.iloc[row_index, 0]
        else:
            data[level_0_name] = feature.iloc[row_index]
    return data


class DataLoader:
    """
    A class that represents a data loader for a dataset. This class is used to load the dataset in batches in a specific data type (jax, pandas, or numpy).

    Parameters
    ----------
    dataset: Dataset
        The dataset to load.
    batch_size: int
        The size of the batch.
    dtype: Literal["jax", "pandas", "numpy"]
        The data type to load the dataset in.

    Example
    -------

    >>> from holisticai.datasets import load_dataset
    >>> dataset = load_dataset("adult")
    >>> dataloader = DataLoader(dataset, batch_size=32, dtype="jax")
    >>> for batch in dataloader:
    ...     print(batch)
    """

    def __init__(self, dataset: Dataset, batch_size: int, dtype: Literal["jax", "pandas", "numpy"]):
        self.batch_size = batch_size
        self.dataset = dataset
        self.dtype = dtype
        self.num_batches = int(np.ceil(len(dataset) / batch_size))

    def batched(self):
        def batch_generator(batch_size):
            for i in range(self.num_batches):
                batch = Dataset(self.dataset.data.iloc[i * batch_size : (i + 1) * batch_size])
                yield batch

        if self.dtype == "jax":
            import jax.numpy as jnp

            def batch_generator_jax(batch_size):
                for batch in batch_generator(batch_size):
                    yield {f: jnp.array(batch[f].values) for f in batch.features}

            return batch_generator_jax(self.batch_size)

        if self.dtype == "pandas":

            def batch_generator_pandas(batch_size):
                for batch in batch_generator(batch_size):
                    yield {f: batch[f] for f in batch.features}

            return batch_generator_pandas(self.batch_size)

        if self.dtype == "numpy":

            def batch_generator_numpy(batch_size):
                for batch in batch_generator(batch_size):
                    yield {f: batch[f].values for f in batch.features}

            return batch_generator_numpy(self.batch_size)
        return batch_generator(self.batch_size)

    def __iter__(self):
        """Iterates over the batches in the dataset."""
        yield from self.batched()

    def _repr_html_(self):
        obj = {
            "dtype": "DataLoader",
            "attributes": {"Number of Batches": self.num_batches, "Batch Size": self.batch_size, "Type": self.dtype},
            "nested_objects": [
                {
                    "dtype": "Dataset",
                    "attributes": {"Number of Rows": self.dataset.num_rows, "Features": self.dataset.features},
                }
            ],
        }
        return generate_html_for_generic_object(obj, feature_columns=5)


class Dataset:
    """Represents a dataset.

    Parameters
    ----------
    data: pd.DataFrame
        The underlying data of the dataset.
    features: list[str]
        The list of features in the dataset.
    num_rows: int
        The number of rows in the dataset.
    random_state: np.random.RandomState
        The random state used for sampling.
    """

    def __update_metadata(self):
        """Updates the metadata of the dataset."""
        self.features = list(self.data.columns.get_level_values(0).unique())
        self.num_rows = len(self.data)
        self.indices = self.data.index

        features_values = self.data.columns.get_level_values("features")
        features_counts = features_values.value_counts()
        self.features_is_series = {key: (value == 1) for key, value in features_counts.items()}

    def __init__(self, _data: pd.DataFrame | None = None, _metadata=None, **kargs):
        if _data is None:
            self.data = {}
            for name, value in kargs.items():
                if isinstance(value, pd.DataFrame):
                    self.data[name] = value.reset_index(drop=True)
                elif isinstance(value, pd.Series):
                    self.data[name] = pd.Series(value.reset_index(drop=True), name=name).astype(value.dtype)
                else:
                    msg = f"Variable '{name}' is of type {type(value)}, but only pd.DataFrame and pd.Series are supported."
                    raise TypeError(msg)
            self.data = pd.concat(self.data, axis=1)
            self.data.columns = self.data.columns.set_names(["features", "subfeatures"])
            self.data.reset_index(drop=True)
        else:
            self.data = _data.reset_index(drop=True)
        self.__update_metadata()
        self._metadata = _metadata
        self.random_state = np.random.RandomState()

    def remove_columns(self, columns: str | list):
        """Returns a new dataset with the given columns removed."""
        return Dataset(self.data.drop(columns, level=0, axis=1), _metadata=self._metadata)

    def rename(self, renames):
        """Returns a new dataset with renamed columns."""
        return Dataset(self.data.rename(columns=renames, level=0), _metadata=self._metadata)

    def select(self, indices: Iterable):
        """Returns a new dataset with selected rows based on the given indices."""
        existing_indices = [idx for idx in indices if idx in self.indices]
        return Dataset(self.data.iloc[existing_indices], _metadata=self._metadata)

    def sample(self, n, random_state=None):
        """Returns a random sample of n rows from the dataset."""
        if random_state is None:
            random_state = self.random_state
        return Dataset(
            sample_n(self.data, n, random_state=random_state).reset_index(drop=True), _metadata=self._metadata
        )

    def filter(self, fn):
        """Returns a new dataset with rows filtered based on the given function."""

        def fnw(row):
            new_row = {k[0] if k[0] == k[1] else k: v for k, v in row.to_dict().items()}
            return fn(new_row)

        new_datad = self.data[self.data.apply(fnw, axis=1)]
        return Dataset(new_datad, _metadata=self._metadata)

    def groupby(self, key: list[str] | str):
        """Returns a new GroupByDataset object based on the given key."""
        if isinstance(key, list):
            key = [(key[0], key[0]), (key[1], key[1])]
        elif isinstance(key, str):
            key = [(key, key)]
        else:
            raise TypeError
        return GroupByDataset(self.data.groupby(key, observed=True))

    def map(self, fn, vectorized=True):
        """Applies a function to the dataset and returns a new dataset.

        Parameters
        ----------

        fn: function
            The function to apply to the dataset.
        vectorized: bool
            Whether to apply the function in a vectorized manner or not.
        """
        if vectorized:

            def fnw(x):
                ds = {level: x.xs(level, axis=1, level="features") for level in x.columns.levels[0]}
                return fn(ds)

            new_data = fnw(self.data)
            updated_data = pd.concat(new_data, axis=1)
            updated_data.columns = pd.MultiIndex.from_tuples(
                [(key, key) for key, serie in new_data.items()], names=["features", "subfeatures"]
            )
        else:

            def fnw(row):
                result = {}
                for upper in row.index.levels[0]:
                    sub_row = row[upper]
                    if isinstance(sub_row, pd.Series):
                        result[upper] = sub_row.to_dict() if len(sub_row) > 1 else sub_row.item()
                    elif isinstance(sub_row, pd.DataFrame):
                        sub_row = sub_row.squeeze()
                        result[upper] = sub_row.to_dict() if len(sub_row) > 1 else sub_row.squeeze().item()
                return fn(result)

            updated_data = self.data.apply(fnw, axis=1, result_type="expand")
            updated_data = pd.DataFrame(updated_data)
            new_columns = pd.MultiIndex.from_tuples(
                [(col, col) for col in updated_data.columns], names=["features", "subfeatures"]
            )
            updated_data.columns = new_columns
        self.data.update(updated_data)
        new_data = pd.concat([self.data, updated_data[updated_data.columns.difference(self.data.columns)]], axis=1)
        return Dataset(new_data, _metadata=self._metadata)

    def train_test_split(self, test_size=0.3, **kargs):
        """Splits the dataset into train and test datasets."""
        train_df, test_df = train_test_split(self.data, test_size=test_size, **kargs)
        train = Dataset(train_df, _metadata=self._metadata)
        test = Dataset(test_df, _metadata=self._metadata)
        return DatasetDict(train=train, test=test)

    def __len__(self):
        """Returns the number of samples in the dataset."""
        return self.num_rows

    def __repr__(self):
        return json.dumps(self.repr_info())

    def repr_info(self):
        return {
            "dtype": "Dataset",
            "attributes": {"Number of Rows": self.num_rows, "Features": [" , ".join(self.features)]},
            "metadata": self._metadata,
        }

    def _repr_html_(self):
        return generate_html_for_generic_object(self.repr_info(), feature_columns=5)

    def __getitem__(self, key: str | int | list):
        """Returns a subset of the dataset based on the given key."""
        if isinstance(key, str):
            feature = self.data.xs(key, level="features", axis=1)
            if feature.shape[1] == 1:
                return feature.iloc[:, 0]
            feature.columns = list(feature.columns)
            return feature

        if isinstance(key, int):
            return dataframe_to_level_dict_with_series(self.data, key)

        if isinstance(key, list) and all(isinstance(k, str) for k in key):
            subset = self.data.loc[:, key]
            subset.columns = subset.columns.droplevel("features")
            return subset

        raise NotImplementedError

    def __setitem__(self, key, value):
        if not isinstance(key, str):
            raise TypeError("Key must be a string.")

        feature_exists = key in self.data.columns.levels[0]
        existing_subfeatures = self.data[key].columns if feature_exists else []
        if feature_exists:
            self.data = self.data.drop(columns=key, level="features")

        if isinstance(value, pd.DataFrame):
            new_columns = pd.MultiIndex.from_product([[key], value.columns])
            value.columns = new_columns
            self.data = self.data.join(value)

        if isinstance(value, pd.Series):
            new_column = (key, value.name or len(existing_subfeatures))
            self.data[new_column] = value

        self.data.columns = self.data.columns.set_names(["features", "subfeatures"])


def concatenate_datasets(part_datasets: list[Dataset]):
    features = part_datasets[0].features
    return Dataset(
        **{feat: pd.concat([p[feat] for p in part_datasets], axis=0).reset_index(drop=True) for feat in features}
    )


def convert_to_pandas(data):
    if all(isinstance(i, (list, tuple)) for i in data):
        return pd.DataFrame(data)
    return pd.Series(data)


def split_dataframe_by_level(df, level=0):
    dataframes = {}
    for key in df.columns.levels[level]:
        dataframes[key] = df.xs(key, axis=1, level=level)
    return dataframes


def apply_fn_to_multilevel_df(df, fn):
    result_df = pd.DataFrame()
    for level in df.columns.levels[0]:
        subset = df.xs(level, axis=1, level=0)
        result = subset.apply(fn, axis=1, result_type="expand")
        result.columns = pd.MultiIndex.from_product([[level], result.columns])
        result_df = pd.concat([result_df, result], axis=1)
    return result_df


def sample_n(group: pd.DataFrame, n: int, random_state: Union[RandomState, None] = None) -> pd.DataFrame:
    if len(group) < n:
        return group
    return group.sample(n=n, replace=False, random_state=random_state)



from __future__ import annotations

import numbers

import numpy as np
import pandas as pd
from joblib import Parallel, delayed
from sklearn.inspection import partial_dependence
from sklearn.metrics import accuracy_score, r2_score

from holisticai.utils._definitions import ModelProxy, PartialDependence


def get_partial_dependence(
    estimator,
    x,
    features,
    sample_weight=None,
    categorical_features=None,
    feature_names=None,
    response_method="auto",
    grid_resolution=100,
    percentiles=(0.05, 0.95),
    method="auto",
    n_jobs=None,
    verbose=0,
    kind="average",
    subsample=1000,
) -> list[dict[str, np.ndarray]]:
    # expand kind to always be a list of str
    kind_ = [kind] * len(features) if isinstance(kind, str) else kind
    if len(kind_) != len(features):
        msg = (
            "When `kind` is provided as a list of strings, it should contain "
            f"as many elements as `features`. `kind` contains {len(kind_)} "
            f"element(s) and `features` contains {len(features)} element(s)."
        )
        raise ValueError(msg)

    if categorical_features is None:
        is_categorical = [(False,) for fxs in features]
    else:
        # we need to create a boolean indicator of which features are
        # categorical from the categorical_features list.
        categorical_features = np.asarray(categorical_features)
        is_categorical = [(categorical_features[fx],) for fx in features]

        for is_cat, kind_plot in zip(is_categorical, kind_):
            if any(is_cat) and kind_plot != "average":
                msg = "It is not possible to display individual effects for" " categorical features."
                raise ValueError(msg)

    if isinstance(subsample, numbers.Integral):
        if subsample <= 0:
            msg = f"When an integer, subsample={subsample} should be positive."
            raise ValueError(msg)

    elif isinstance(subsample, numbers.Real) and (subsample <= 0 or subsample >= 1):
        msg = f"When a floating-point, subsample={subsample} should be in " "the (0, 1) range."
        raise ValueError(msg)

    # compute predictions and/or averaged predictions
    return Parallel(n_jobs=n_jobs, verbose=verbose)(
        delayed(partial_dependence)(
            estimator,
            x,
            fxs,
            sample_weight=sample_weight,
            feature_names=feature_names,
            categorical_features=categorical_features,
            response_method=response_method,
            method=method,
            grid_resolution=grid_resolution,
            percentiles=percentiles,
            kind=kind_plot,
        )  # type: ignore
        for kind_plot, fxs in zip(kind_, features)
    )


class SKLearnModel:
    def __init__(self, predict, predict_proba, fit, score, classes, estimator_type):
        self.predict = predict
        self.fit = fit
        self.score = score
        self.foo_ = "foo_"  # necessary for partial_dependency_function validation
        self._estimator_type = estimator_type
        if predict_proba is not None:
            self.predict_proba = predict_proba
        if classes is not None:
            self.classes_ = classes

    def __sklearn_is_fitted__(self):
        return True


def wrap_sklearn_binary_model(classes: list, predict: callable, predict_proba: callable | None = None):
    def fit(x, y):  # noqa: ARG001
        pass

    def score(x, y):
        pred = predict(x)
        return accuracy_score(y_true=y, y_pred=pred)

    def predict_proba_fn(x):
        if predict_proba is None:
            return None
        prob = predict_proba(x)
        return np.stack([1 - prob, prob], axis=1)

    return SKLearnModel(
        predict=predict,
        predict_proba=predict_proba_fn,
        fit=fit,
        score=score,
        classes=classes,
        estimator_type="classifier",
    )


def wrap_sklearn_multi_classification_model(classes: list, predict: callable, predict_proba: callable | None = None):
    def fit(x, y):  # noqa: ARG001
        pass

    def score(x, y):
        pred = predict(x)
        return accuracy_score(y_true=y, y_pred=pred)

    def predict_proba_fn(x):
        return None if predict_proba is None else predict_proba(x)

    return SKLearnModel(
        predict=predict,
        predict_proba=predict_proba_fn,
        fit=fit,
        score=score,
        classes=classes,
        estimator_type="classifier",
    )


def wrap_sklearn_regression_model(predict: callable):
    def fit(x, y):  # noqa: ARG001
        pass

    def score(x, y):
        pred = predict(x)
        return r2_score(y_true=y, y_pred=pred)

    return SKLearnModel(
        predict=predict,
        predict_proba=None,
        fit=fit,
        score=score,
        classes=None,
        estimator_type="regressor",
    )


def wrap_sklearn_model(proxy: ModelProxy):
    learning_task = proxy.learning_task
    if learning_task == "binary_classification":
        return wrap_sklearn_binary_model(
            classes=proxy.classes,
            predict=proxy.predict,
            predict_proba=proxy.predict_proba,
        )
    if learning_task == "regression":
        return wrap_sklearn_regression_model(predict=proxy.predict)
    if learning_task == "multi_classification":
        return wrap_sklearn_multi_classification_model(
            classes=proxy.classes,
            predict=proxy.predict,
            predict_proba=proxy.predict_proba,
        )
    return None


def compute_partial_dependence(X: pd.DataFrame, features: list[str], proxy: ModelProxy) -> PartialDependence:
    supported_learning_tasks = ["binary_classification", "regression", "multi_classification"]
    if proxy.learning_task not in supported_learning_tasks:
        raise ValueError(f"Learning task {proxy.learning_task} is not supported for partial dependence computation")

    model = wrap_sklearn_model(proxy)
    feature_names = np.array(X.columns)
    method = "auto"
    response_method = "auto"
    grid_resolution = 50
    feature_index = [np.where(feature_names == f)[0][0] for f in features]
    percentiles = (0.05, 0.95) if proxy.learning_task == "regression" else (0, 1)
    partial_dependence = get_partial_dependence(
        model,  # type: ignore
        X,
        features=feature_index,
        kind="both",
        method=method,
        response_method=response_method,
        grid_resolution=grid_resolution,
        percentiles=percentiles,
        n_jobs=1,
    )

    if proxy.learning_task in ["binary_classification", "multi_classification"]:
        nb_classes = len(proxy.classes)
        new_partial_dependence = []
        for c in range(nb_classes):
            part_dep_feat = []
            for p in partial_dependence:
                assert nb_classes == len(p["individual"])
                newp = p.copy()
                newp["individual"] = p["individual"][c][np.newaxis]
                newp["average"] = p["average"][c][np.newaxis]
                part_dep_feat.append(newp)
            new_partial_dependence.append(part_dep_feat)
        return PartialDependence(values=new_partial_dependence)
    return PartialDependence(values=[partial_dependence])



"""
This module implements the zeroth-order optimization attack `ZooAttack`. This is a black-box attack. This attack is a
variant of the Carlini and Wagner attack which uses ADAM coordinate descent to perform numerical estimation of
gradients.

| Paper link: https://arxiv.org/abs/1708.03999
"""

from __future__ import annotations

from typing import Any, Optional

import numpy as np
import pandas as pd
from holisticai.robustness.attackers.classification.commons import (
    format_function_predict_proba,
    to_categorical,
    x_array_to_df,
    x_to_nd_array,
)
from scipy.ndimage import zoom

BATCH_SIZE = 1


class ZooAttack:
    """
    The black-box zeroth-order optimization attack from Pin-Yu Chen et al. (2018). This attack is a variant of the
    C&W attack which uses ADAM coordinate descent to perform numerical estimation of gradients.

    Parameters
    ----------
    name : str, optional
        The name of the attack. The default is "Zoo".
    confidence : float, optional
        Confidence of adversarial examples. A higher value produces examples that are farther away, but more strongly\\
        classified as adversarial. The default is 0.0.
    targeted : bool, optional
        Indicates whether the attack is targeted. The default is False. If True, the positive ground truth is used as the target.
    learning_rate : float, optional
        The learning rate for the ADAM optimizer. The default is 1e-2.
    max_iter : int, optional
        The maximum number of iterations. The default is 20.
    binary_search_steps : int, optional
        The number of binary search steps. The default is 10.
    initial_const : float, optional
        The initial constant used to scale the adversarial perturbation. The default is 1e-3.
    abort_early : bool, optional
        Indicates whether to abort the optimization early. The default is True.
    use_resize : bool, optional
        Indicates whether to use resizing. The default is False.
    use_importance : bool, optional
        Indicates whether to use importance sampling. The default is False.
    nb_parallel : int, optional
        The number of parallel coordinates to update. The default is 1.
    variable_h : float, optional
        The variable h. The default is 0.2.
    verbose : bool, optional
        Indicates whether to print verbose output. The default is True.
    input_is_feature_vector : bool, optional
        Indicates whether the input is a feature vector. The default is False.
    proxy : callable, optional
        The model used to predict the probabilities of the input. The default is None.
    input_size : int, optional
        The size of the input. The default is 0.
    nb_classes : int, optional
        The number of classes. The default is 2.
    adam_mean : Optional[NDArray|ArrayLike|None], optional
        The mean of the ADAM optimizer. The default is None.
    adam_var : Optional[NDArray|ArrayLike|None], optional
        The variance of the ADAM optimizer. The default is None.
    adam_epochs : Optional[NDArray|ArrayLike|None], optional
        The epochs of the ADAM optimizer. The default is None.
    """

    def __init__(
        self,
        name="Zoo",
        confidence=0.0,
        targeted=False,
        learning_rate=1e-2,
        max_iter=20,
        binary_search_steps=10,
        initial_const=1e-3,
        abort_early=True,
        use_resize=False,
        use_importance=False,
        nb_parallel=1,
        variable_h=0.2,
        verbose=True,
        input_is_feature_vector=False,
        proxy=None,
        input_size=0,
        nb_classes=2,
        adam_mean=None,
        adam_var=None,
        adam_epochs=None,
    ):
        self.name = name
        self.confidence = confidence
        self.targeted = targeted
        self.learning_rate = learning_rate
        self.max_iter = max_iter
        self.binary_search_steps = binary_search_steps
        self.initial_const = initial_const
        self.abort_early = abort_early
        self.use_resize = use_resize
        self.use_importance = use_importance
        self.nb_parallel = nb_parallel
        self.batch_size = BATCH_SIZE
        self.variable_h = variable_h
        self.verbose = verbose
        self.input_is_feature_vector = input_is_feature_vector
        self.predict_proba_fn = format_function_predict_proba(proxy.learning_task, proxy.predict_proba)
        self.input_size = input_size
        self.nb_classes = nb_classes
        self.adam_mean = adam_mean
        self.adam_var = adam_var
        self.adam_epochs = adam_epochs

    def _initialize_vars(self, x: np.ndarray) -> None:
        """
        Initialize the variables.

        Parameters
        ----------
        x : np.ndarray
            The input samples.
        """
        self.input_shape = tuple(x.shape[1:])
        self.input_size = np.prod(self.input_shape)
        if len(self.input_shape) == 1:
            self.input_is_feature_vector = True
            if self.batch_size != 1:
                raise ValueError(
                    "The current implementation of Zeroth-Order Optimisation attack only supports "
                    "`batch_size=1` with feature vectors as input."
                )
        else:
            self.input_is_feature_vector = False

        # Initialize some internal variables
        self._init_size = 32
        if self.abort_early:
            self._early_stop_iters = self.max_iter // 10 if self.max_iter >= 10 else self.max_iter

        # Initialize noise variable to zero
        if self.input_is_feature_vector:
            self.use_resize = False
            self.use_importance = False

        if self.use_resize:
            dims = (self.batch_size, self.input_shape[0], self._init_size, self._init_size)
            self._current_noise = np.zeros(dims, dtype=np.float32)
        else:
            self._current_noise = np.zeros((self.batch_size, *self.input_shape), dtype=np.float32)
        self._sample_prob = np.ones(self._current_noise.size, dtype=np.float32) / self._current_noise.size

    def _loss(
        self, x: np.ndarray, x_adv: np.ndarray, target: np.ndarray, c_weight: np.ndarray
    ) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
        """
        Compute the loss function values.

        Parameters
        ----------
        x : np.ndarray
            The original input.
        x_adv : np.ndarray
            The adversarial input.
        target : np.ndarray
            The target values.
        c_weight : np.ndarray
            The weight of the constant.

        Returns
        -------
        tuple[np.ndarray, np.ndarray, np.ndarray]
            The predictions, the L2 distances, and the loss values.
        """
        l2dist = np.sum(np.square(x - x_adv).reshape(x_adv.shape[0], -1), axis=1)
        ratios = [1.0] + [int(new_size) / int(old_size) for new_size, old_size in zip(self.input_shape, x.shape[1:])]
        preds = self.predict_proba(np.array(zoom(x_adv, zoom=ratios)))
        z_target = np.sum(preds * target, axis=1)
        z_other = np.max(
            preds * (1 - target) + (np.min(preds, axis=1) - 1)[:, np.newaxis] * target,
            axis=1,
        )

        if self.targeted:
            # If targeted, optimize for making the target class most likely
            loss = np.maximum(z_other - z_target + self.confidence, 0)
        else:
            # If untargeted, optimize for making any other class most likely
            loss = np.maximum(z_target - z_other + self.confidence, 0)

        return preds, l2dist, c_weight * loss + l2dist

    def generate(self, x_df: pd.DataFrame, y: Optional[np.ndarray] = None) -> pd.DataFrame:
        """
        Generate adversarial samples and return them in an array.

        Parameters
        ----------
        x_df : pd.DataFrame
            The input samples.
        y : Optional[np.ndarray], optional
            The target labels. The default is None.

        Returns
        -------
        pd.DataFrame
            The adversarial samples.
        """
        self._initialize_vars(x_df)
        feature_names = list(x_df.columns)
        self.predict_proba = lambda x: self.predict_proba_fn(x, feature_names)

        x = x_to_nd_array(x_df)

        self._clip_values = (np.min(x), np.max(x))

        if y is not None:
            y = to_categorical(y, nb_classes=self.nb_classes)

        # Check that `y` is provided for targeted attacks
        if self.targeted and y is None:  # pragma: no cover
            raise ValueError("Target labels `y` need to be provided for a targeted attack.")

        # No labels provided, use model prediction as correct class
        if y is None:
            y = self.predict_proba(x)

        if self.nb_classes == 2 and y.shape[1] == 1:  # pragma: no cover
            raise ValueError(
                "This attack has not yet been tested for binary classification with a single output classifier."
            )

        # Compute adversarial examples with implicit batching
        nb_batches = int(np.ceil(x.shape[0] / float(self.batch_size)))
        x_adv_list = []
        for batch_id in range(nb_batches):
            batch_index_1, batch_index_2 = batch_id * self.batch_size, (batch_id + 1) * self.batch_size
            x_batch = x[batch_index_1:batch_index_2]
            y_batch = y[batch_index_1:batch_index_2]
            res = self._generate_batch(x_batch, y_batch)
            x_adv_list.append(res)
        x_adv = np.vstack(x_adv_list)

        # Apply clip
        # clip_min, clip_max = self._clip_values
        # np.clip(x_adv, clip_min-0.1, clip_max+0.1, out=x_adv)

        return x_array_to_df(x_adv, feature_names=feature_names)

    def _generate_batch(self, x_batch: np.ndarray, y_batch: np.ndarray) -> np.ndarray:
        """
        Run the attack on a batch of images and labels.

        Parameters
        ----------
        x_batch : np.ndarray
            A batch of original examples.
        y_batch : np.ndarray
            A batch of targets (0-1 hot).

        Returns
        -------
        np.ndarray
            A batch of adversarial examples.
        """
        # Initialize binary search
        c_current = self.initial_const * np.ones(x_batch.shape[0])
        c_lower_bound = np.zeros(x_batch.shape[0])
        c_upper_bound = 1e10 * np.ones(x_batch.shape[0])

        # Initialize best distortions and best attacks globally
        o_best_dist = np.inf * np.ones(x_batch.shape[0])
        o_best_attack = x_batch.copy()

        # Start with a binary search
        for _ in range(self.binary_search_steps):
            # Run with 1 specific binary search step
            best_dist, best_label, best_attack = self._generate_bss(x_batch, y_batch, c_current)

            # Update best results so far
            o_best_attack[best_dist < o_best_dist] = best_attack[best_dist < o_best_dist]
            o_best_dist[best_dist < o_best_dist] = best_dist[best_dist < o_best_dist]

            # Adjust the constant as needed
            c_current, c_lower_bound, c_upper_bound = self._update_const(
                y_batch, best_label, c_current, c_lower_bound, c_upper_bound
            )

        return o_best_attack

    def _update_const(
        self,
        y_batch: np.ndarray,
        best_label: np.ndarray,
        c_batch: np.ndarray,
        c_lower_bound: np.ndarray,
        c_upper_bound: np.ndarray,
    ) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
        """
        Update constant `c_batch` from the ZOO objective. This characterizes the trade-off between attack strength and
        amount of noise introduced.

        Parameters
        ----------
        y_batch : np.ndarray
            A batch of targets (0-1 hot).
        best_label : np.ndarray
            The best labels.
        c_batch : np.ndarray
            A batch of constants.
        c_lower_bound : np.ndarray
            The lower bound of the constant.
        c_upper_bound : np.ndarray
            The upper bound of the constant.

        Returns
        -------
        tuple[np.ndarray, np.ndarray, np.ndarray]
            The updated constant, lower bound, and upper bound.
        """

        comparison = [
            self._compare(best_label[i], np.argmax(y_batch[i])) and best_label[i] != -np.inf
            for i in range(len(c_batch))
        ]
        for i, comp in enumerate(comparison):
            if comp:
                # Successful attack
                c_upper_bound[i] = min(c_upper_bound[i], c_batch[i])
                if c_upper_bound[i] < 1e9:
                    c_batch[i] = (c_lower_bound[i] + c_upper_bound[i]) / 2
            else:
                # Failure attack
                c_lower_bound[i] = max(c_lower_bound[i], c_batch[i])
                c_batch[i] = (c_lower_bound[i] + c_upper_bound[i]) / 2 if c_upper_bound[i] < 1e9 else c_batch[i] * 10

        return c_batch, c_lower_bound, c_upper_bound

    def _compare(self, object1: Any, object2: Any) -> bool:
        """
        Check two objects for equality if the attack is targeted, otherwise check for inequality.

        Parameters
        ----------
        object1 : Any
            The first object.
        object2 : Any
            The second object.

        Returns
        -------
        bool
            The result of the comparison.
        """
        return object1 == object2 if self.targeted else object1 != object2

    def _generate_bss(
        self, x_batch: np.ndarray, y_batch: np.ndarray, c_batch: np.ndarray
    ) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
        """
        Generate adversarial examples for a batch of inputs with a specific batch of constants.

        Parameters
        ----------
        x_batch : np.ndarray
            A batch of original examples.
        y_batch : np.ndarray
            A batch of targets (0-1 hot).
        c_batch : np.ndarray
            A batch of constants.

        Returns
        -------
        tuple[np.ndarray, np.ndarray, np.ndarray]
            The best distortions, the best labels, and the best attacks.
        """

        x_orig = x_batch.astype(np.float32)
        fine_tuning = np.full(x_batch.shape[0], False, dtype=bool)
        prev_loss = 1e6 * np.ones(x_batch.shape[0])
        prev_l2dist = np.zeros(x_batch.shape[0])

        # Resize and initialize Adam
        if self.use_resize:
            x_orig = self._resize_image(x_orig, self._init_size, self._init_size, True)
            assert (x_orig != 0).any()
            x_adv = x_orig.copy()
        else:
            x_orig = x_batch
            self._reset_adam(np.prod(self.input_shape).item())
            if x_batch.shape == self._current_noise.shape:
                self._current_noise.fill(0)
            else:
                self._current_noise = np.zeros(x_batch.shape, dtype=np.float32)
            x_adv = x_orig.copy()

        # Initialize best distortions, best changed labels and best attacks
        best_dist = np.inf * np.ones(x_adv.shape[0])
        best_label = -np.inf * np.ones(x_adv.shape[0])
        best_attack = np.array([x_adv[i] for i in range(x_adv.shape[0])])

        for iter_ in range(self.max_iter):
            # Upscaling for very large number of iterations
            if self.use_resize:
                if iter_ == 2000:
                    x_adv = self._resize_image(x_adv, 64, 64)
                    x_orig = zoom(
                        x_orig,
                        [
                            1,
                            x_adv.shape[1] / x_orig.shape[1],
                            x_adv.shape[2] / x_orig.shape[2],
                            x_adv.shape[3] / x_orig.shape[3],
                        ],
                    )
                elif iter_ == 10000:
                    x_adv = self._resize_image(x_adv, 128, 128)
                    x_orig = zoom(
                        x_orig,
                        [
                            1,
                            x_adv.shape[1] / x_orig.shape[1],
                            x_adv.shape[2] / x_orig.shape[2],
                            x_adv.shape[3] / x_orig.shape[3],
                        ],
                    )

            # Compute adversarial examples and loss
            x_adv = self._optimizer(x_adv, y_batch, c_batch)
            preds, l2dist, loss = self._loss(x_orig, x_adv, y_batch, c_batch)

            # Reset Adam if a valid example has been found to avoid overshoot
            mask_fine_tune = (~fine_tuning) & (loss == l2dist) & (prev_loss != prev_l2dist)
            fine_tuning[mask_fine_tune] = True
            self._reset_adam(self.adam_mean.size, np.repeat(mask_fine_tune, x_adv[0].size))  # type: ignore
            prev_l2dist = l2dist

            # Abort early if no improvement is obtained
            if self.abort_early and iter_ % self._early_stop_iters == 0:
                if (loss > 0.9999 * prev_loss).all():
                    break
                prev_loss = loss

            # Adjust the best result
            labels_batch = np.argmax(y_batch, axis=1)
            for i, (dist, pred) in enumerate(zip(l2dist, np.argmax(preds, axis=1))):
                if dist < best_dist[i] and self._compare(pred, labels_batch[i]):
                    best_dist[i] = dist
                    best_attack[i] = x_adv[i]
                    best_label[i] = pred

        # Resize images to original size before returning
        best_attack = np.array(best_attack)
        if self.use_resize:
            if not self.channels_first:
                best_attack = zoom(
                    best_attack,
                    [
                        1,
                        int(x_batch.shape[1]) / best_attack.shape[1],
                        int(x_batch.shape[2]) / best_attack.shape[2],
                        1,
                    ],
                )
            else:
                best_attack = zoom(
                    best_attack,
                    [
                        1,
                        1,
                        int(x_batch.shape[2]) / best_attack.shape[2],
                        int(x_batch.shape[2]) / best_attack.shape[3],
                    ],
                )

        return best_dist, best_label, best_attack

    def _optimizer(self, x: np.ndarray, targets: np.ndarray, c_batch: np.ndarray) -> np.ndarray:
        """
        Run the ADAM optimizer for a batch of inputs.

        Parameters
        ----------
        x : np.ndarray
            A batch of original examples.
        targets : np.ndarray
            A batch of targets (0-1 hot).
        c_batch : np.ndarray
            A batch of constants.

        Returns
        -------
        np.ndarray
            A batch of adversarial examples.
        """
        # Variation of input for computing loss, same as in original implementation
        coord_batch = np.repeat(self._current_noise, 2 * self.nb_parallel, axis=0)
        coord_batch = coord_batch.reshape(2 * self.nb_parallel * self._current_noise.shape[0], -1)

        # Sample indices to prioritize for optimization
        if self.use_importance and np.unique(self._sample_prob).size != 1:
            indices = (
                np.random.choice(
                    coord_batch.shape[-1] * x.shape[0],
                    self.nb_parallel * self._current_noise.shape[0],
                    replace=False,
                    p=self._sample_prob.flatten(),
                )
                % coord_batch.shape[-1]
            )
        else:
            try:
                indices = (
                    np.random.choice(
                        coord_batch.shape[-1] * x.shape[0],
                        self.nb_parallel * self._current_noise.shape[0],
                        replace=False,
                    )
                    % coord_batch.shape[-1]
                )
            except ValueError as error:  # pragma: no cover
                if "Cannot take a larger sample than population when 'replace=False'" in str(error):
                    raise ValueError(
                        "Too many samples are requested for the random indices. Try to reduce the number of parallel"
                        "coordinate updates `nb_parallel`."
                    ) from error

                raise

        # Create the batch of modifications to run
        for i in range(self.nb_parallel * self._current_noise.shape[0]):
            coord_batch[2 * i, indices[i]] += self.variable_h
            coord_batch[2 * i + 1, indices[i]] -= self.variable_h

        # Compute loss for all samples and coordinates, then optimize
        expanded_x = np.repeat(x, 2 * self.nb_parallel, axis=0).reshape((-1,) + x.shape[1:])
        expanded_targets = np.repeat(targets, 2 * self.nb_parallel, axis=0).reshape((-1,) + targets.shape[1:])
        expanded_c = np.repeat(c_batch, 2 * self.nb_parallel)
        _, _, loss = self._loss(
            expanded_x,
            expanded_x + coord_batch.reshape(expanded_x.shape),
            expanded_targets,
            expanded_c,
        )
        if self.adam_mean is not None and self.adam_var is not None and self.adam_epochs is not None:
            self._current_noise = self._optimizer_adam_coordinate(
                loss,
                indices,
                self.adam_mean,
                self.adam_var,
                self._current_noise,
                self.learning_rate,
                self.adam_epochs,
                True,
            )
        else:
            raise ValueError("Unexpected `None` in `adam_mean`, `adam_var` or `adam_epochs` detected.")

        if self.use_importance and self._current_noise.shape[2] > self._init_size:
            self._sample_prob = self._get_prob(self._current_noise).flatten()

        return x + self._current_noise

    def _optimizer_adam_coordinate(
        self,
        losses: np.ndarray,
        index: np.ndarray,
        mean: np.ndarray,
        var: np.ndarray,
        current_noise: np.ndarray,
        learning_rate: float,
        adam_epochs: np.ndarray,
        proj: bool,
    ) -> np.ndarray:
        """
        Implementation of the ADAM optimizer for coordinate descent.

        Parameters
        ----------
        losses : np.ndarray
            Overall loss.
        index : np.ndarray
            Indices of the coordinates to update.
        mean : np.ndarray
            The mean of the gradient (first moment).
        var : np.ndarray
            The uncentered variance of the gradient (second moment).
        current_noise : np.ndarray
            The current noise.
        learning_rate : float
            Learning rate for Adam optimizer.
        adam_epochs : np.ndarray
            Epochs to run the Adam optimizer.
        proj : bool
            Whether to project the noise to the L_p ball.

        Returns
        -------
        np.ndarray
            Updated noise for coordinate descent.
        """
        beta1, beta2 = 0.9, 0.999

        # Estimate grads from loss variation (constant `h` from the paper is fixed to .0001)
        grads = np.array([(losses[i] - losses[i + 1]) / (2 * self.variable_h) for i in range(0, len(losses), 2)])

        # ADAM update
        mean[index] = beta1 * mean[index] + (1 - beta1) * grads
        var[index] = beta2 * var[index] + (1 - beta2) * grads**2

        corr = (np.sqrt(1 - np.power(beta2, adam_epochs[index]))) / (1 - np.power(beta1, adam_epochs[index]))
        orig_shape = current_noise.shape
        current_noise = current_noise.reshape(-1)
        current_noise[index] -= learning_rate * corr * mean[index] / (np.sqrt(var[index]) + 1e-8)
        adam_epochs[index] += 1

        if proj and hasattr(self, "_clip_values") and self._clip_values is not None:
            clip_min, clip_max = self._clip_values
            current_noise[index] = np.clip(current_noise[index], clip_min, clip_max)

        return current_noise.reshape(orig_shape)

    def _reset_adam(self, nb_vars: int, indices: Optional[np.ndarray] = None) -> None:
        """
        Reset the ADAM optimizer.

        Parameters
        ----------
        nb_vars : int
            The number of variables.
        indices : Optional[np.ndarray], optional
            The indices to reset. The default is None.
        """
        # If variables are already there and at the right size, reset values
        if self.adam_mean is not None and self.adam_mean.size == nb_vars:
            if indices is None:
                self.adam_mean.fill(0)
                self.adam_var.fill(0)  # type: ignore
                self.adam_epochs.fill(1)  # type: ignore
            else:
                self.adam_mean[indices] = 0
                self.adam_var[indices] = 0  # type: ignore
                self.adam_epochs[indices] = 1  # type: ignore
        else:
            # Allocate Adam variables
            self.adam_mean = np.zeros(nb_vars, dtype=np.float32)
            self.adam_var = np.zeros(nb_vars, dtype=np.float32)
            self.adam_epochs = np.ones(nb_vars, dtype=int)

    def _resize_image(self, x: np.ndarray, size_x: int, size_y: int, reset: bool = False) -> np.ndarray:
        """
        Resize the image to a specific size.

        Parameters
        ----------
        x : np.ndarray
            The input image.
        size_x : int
            The size in the x direction.
        size_y : int
            The size in the y direction.
        reset : bool, optional
            Indicates whether to reset the image. The default is False.

        Returns
        -------
        np.ndarray
            The resized image.
        """
        if not self.channels_first:
            dims = (x.shape[0], size_x, size_y, x.shape[-1])
        else:
            dims = (x.shape[0], x.shape[1], size_x, size_y)
        nb_vars = np.prod(dims).item()

        if reset:
            # Reset variables to original size and value
            if dims == x.shape:
                resized_x = x
                if x.shape == self._current_noise.shape:
                    self._current_noise.fill(0)
                else:
                    self._current_noise = np.zeros(x.shape, dtype=np.float32)
            else:
                resized_x = zoom(
                    x,
                    (
                        1,
                        dims[1] / x.shape[1],
                        dims[2] / x.shape[2],
                        dims[3] / x.shape[3],
                    ),
                )
                self._current_noise = np.zeros(dims, dtype=np.float32)
            self._sample_prob = np.ones(nb_vars, dtype=np.float32) / nb_vars
        else:
            # Rescale variables and reset values
            resized_x = zoom(x, (1, dims[1] / x.shape[1], dims[2] / x.shape[2], dims[3] / x.shape[3]))
            self._sample_prob = self._get_prob(self._current_noise, double=True).flatten()
            self._current_noise = np.zeros(dims, dtype=np.float32)

        # Reset Adam
        self._reset_adam(nb_vars)

        return resized_x

    def _get_prob(self, prev_noise: np.ndarray, double: bool = False) -> np.ndarray:
        """
        Compute the probability of each pixel to be selected for optimization.

        Parameters
        ----------
        prev_noise : np.ndarray
            The previous noise.
        double : bool, optional
            Indicates whether to double the size. The default is False.

        Returns
        -------
        np.ndarray
            The probability of each pixel to be selected for optimization.
        """
        dims = list(prev_noise.shape)
        channel_index = 1 if self.channels_first else 3

        # Double size if needed
        if double:
            dims = [2 * size if i not in [0, channel_index] else size for i, size in enumerate(dims)]

        prob = np.empty(shape=dims, dtype=np.float32)
        image = np.abs(prev_noise)

        for channel in range(prev_noise.shape[channel_index]):
            if not self.channels_first:
                image_pool = self._max_pooling(image[:, :, :, channel], dims[1] // 8)
                if double:
                    prob[:, :, :, channel] = np.abs(zoom(image_pool, [1, 2, 2]))
                else:
                    prob[:, :, :, channel] = image_pool
            elif self.channels_first:
                image_pool = self._max_pooling(image[:, channel, :, :], dims[2] // 8)
                if double:
                    prob[:, channel, :, :] = np.abs(zoom(image_pool, [1, 2, 2]))
                else:
                    prob[:, channel, :, :] = image_pool

        prob /= np.sum(prob)

        return prob

    @staticmethod
    def _max_pooling(image: np.ndarray, kernel_size: int) -> np.ndarray:
        """
        Perform max pooling on the image.

        Parameters
        ----------
        image : np.ndarray
            The input image.
        kernel_size : int
            The size of the kernel.

        Returns
        -------
        np.ndarray
            The pooled image.
        """
        img_pool = np.copy(image)
        for i in range(0, image.shape[1], kernel_size):
            for j in range(0, image.shape[2], kernel_size):
                img_pool[:, i : i + kernel_size, j : j + kernel_size] = np.max(
                    image[:, i : i + kernel_size, j : j + kernel_size],
                    axis=(1, 2),
                    keepdims=True,
                )

        return img_pool



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
# z
# import os
# import sys
# sys.path.insert(0, os.path.abspath('.'))

import sys
import os
import shutil

sys.path.insert(0, os.path.abspath('.'))
import utils.xai_image_plots as xai_utils
import inspect

os.makedirs('_static/images', exist_ok=True)

for name, obj in inspect.getmembers(xai_utils):
    if inspect.isfunction(obj) and name.startswith('image_'):
        obj()



# -- Project information -----------------------------------------------------

project = "holisticai"
copyright = "2024, Holistic AI"
author = "Holistic AI"


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.

extensions = [
    "myst_parser",
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.coverage",
    "sphinx.ext.napoleon",
    'sphinx.ext.viewcode',
    "sphinx.ext.doctest",
    "sphinx.ext.extlinks",
    "sphinx.ext.intersphinx",
    #"sphinx.ext.linkcode",
    "sphinx.ext.mathjax",
    #"sphinx_gallery.gen_gallery",
    #"sphinx_autodoc_typehints",
    "nbsphinx",
    "sphinx_copybutton",
    "sphinx_design",
    #"sphinx_prompt",
    #"numpydoc",
    "matplotlib.sphinxext.plot_directive",
    "sphinx_togglebutton",
    "sphinxcontrib.youtube",
]

intersphinx_mapping = {
    "python3": ("https://docs.python.org/3", None),
    "numpy": ("https://numpy.org/doc/stable/", None),
    "pandas": ("https://pandas.pydata.org/pandas-docs/stable/", None),
    "sklearn": ("https://scikit-learn.org/stable/", None),
    "matplotlib": ("https://matplotlib.org/stable/", None)
}

source_suffix = {
    '.rst': 'restructuredtext',
    '.md': 'markdown',
}

nbsphinx_allow_errors = True  # Permitir errores en los notebooks
nbsphinx_execute = 'never'  # Puede ser 'auto', 'always', o 'never'

html_show_sourcelink = False
# autodoc options
autodoc_default_options = {"members": True, "inherited-members": False}

# Turn on autosummary
autosummary_generate = True

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = [
    "_build",
    "Thumbs.db",
    ".DS_Store",
    "generated/*",
    ".ipynb_checkpoints",
]


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
numfig=True

html_logo = "hai_logo.svg"
html_favicon = "holistic_ai.png"

html_theme = "pydata_sphinx_theme"
# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]

# Theme options
html_theme_options = {
    "logo": {"image_dark": "https://assets-global.website-files.com/6305e5d42c283515c3e71b8c/63d771efd50a073bd66193f0_Holistic-AI-Logo-Horizontal-Dark.svg"},
    "twitter_url": "https://twitter.com/holistic_ai",
    "show_version_warning_banner": True,
    "secondary_sidebar_items": [],
#    "announcement": "Visit our website and <a href='https://www.holisticai.com/demo'>schedule a demo</a> with our experts to find out how Holistic AI can help you shield against AI risks.",
    "icon_links": [
        {
            "name": "GitHub",
            "url": "https://github.com/holistic-ai/holisticai",
            "icon": "fa-brands fa-github",
        },
        {
            "name": "Community",
            "url": "https://join.slack.com/t/holisticaicommunity/shared_invite/zt-2jamouyrn-BrMfeoBZIHT8HbLzB3P9QQ",  # required
            "icon": "fa-brands fa-slack",
            "type": "fontawesome",
        }
   ],
}

import os
import shutil
import sys

#sys.path.insert(0, os.path.abspath('../src'))

sys.path.insert(0, os.path.abspath(".."))


# Custom css
html_css_files = [
    "css/custom_style.css",
]

# Custom section headers
napoleon_custom_sections = ["Interpretation", "Description", "Parameters", "Methods", "Returns"]


import numpy as np
from holisticai.bias.mitigation.commons.disparate_impact_remover._numerical_repairer import (
    NumericalRepairer,
)
from holisticai.utils.transformers.bias import BMPreprocessing as BMPre
from holisticai.utils.transformers.bias import SensitiveGroups


class DisparateImpactRemover(BMPre):
    """
    Disparate impact remover [1]_ edits feature values to increase group fairness
    while preserving rank-ordering within groups.

    Parameters
    ----------
    repair_level : float, optional
        The amount of repair to be applied. It should be between 0.0 and 1.0. Default is 1.

    References
    ----------
    .. [1] Feldman, Michael, et al. "Certifying and removing disparate impact."
        proceedings of the 21th ACM SIGKDD international conference on knowledge
        discovery and data mining. 2015.
    """

    def __init__(self, repair_level=1.0):
        self._assert_parameters(repair_level)
        self.repair_level = repair_level
        self._sensgroups = SensitiveGroups()

    def _assert_parameters(self, repair_level):
        """
        Assert parameters

        Parameters
        ----------
        repair_level : float
            The amount of repair to be applied

        Raises
        ------
        ValueError
            If repair_level is not between 0.0 and 1.0
        """
        if not 0.0 <= repair_level <= 1.0:
            raise ValueError("'repair_level' must be between 0.0 and 1.0.")

    def repair_data(self, X, group_a, group_b):
        """
        Repair data to a fair representation

        Parameters
        ----------
        X : array-like
            Input data
        group_a : array-like
            mask vector
        group_b : array-like
            mask vector

        Returns
        -------
        array-like
            Repaired input data matrix
        """
        # Combine the two group masks into a single matrix
        sensitive_features = np.c_[group_a, group_b]

        # Convert the sensitive feature matrix to a numeric representation
        p_attr = self._sensgroups.fit_transform(sensitive_features, convert_numeric=True).to_numpy()

        # Combine the sensitive feature matrix with the input data matrix
        data = np.c_[p_attr, X].tolist()

        # Apply numerical repair to the first column of the combined matrix
        repairer = NumericalRepairer(feature_to_repair=0, repair_level=self.repair_level, kdd=False)
        new_data_matrix_np = repairer.repair(data)

        # Return only the repaired input data matrix (without the sensitive feature column)
        return np.array([np.array(row[1:]) for row in new_data_matrix_np])

    def transform(self, X: np.ndarray, group_a: np.ndarray, group_b: np.ndarray):
        """
        Transform data to a fair representation

        Parameters
        ----------
        X : array-like
            Input data

        group_a : array-like
            mask vector

        group_b : array-like
            mask vector

        Returns
        -------
            Self
        """
        params = self._load_data(X=X, group_a=group_a, group_b=group_b)
        x = params["X"]
        group_a = params["group_a"]
        group_b = params["group_b"]

        return self.repair_data(x, group_a, group_b)

    def fit(self):
        """
        Fit the model

        Returns
        -------
            Self
        """
        return self

    def fit_transform(self, X: np.ndarray, group_a: np.ndarray, group_b: np.ndarray):
        """
        Fit the model and transform data

        Parameters
        ----------
        X : array-like
            Input data
        group_a : array-like
            mask vector
        group_b : array-like
            mask vector

        Returns
        -------
        array-like
            Repaired input data matrix
        """
        return self.fit().transform(X, group_a, group_b)



# Base Imports
import warnings

import numpy as np
import pandas as pd

# utils
from holisticai.utils._validation import (
    _array_like_to_numpy,
    _check_binary,
    _classification_checks,
    _matrix_like_to_numpy,
    _regression_checks,
)

# Efficacy metrics
from sklearn.metrics import accuracy_score, confusion_matrix, roc_auc_score
from sklearn.neighbors import NearestNeighbors


def _group_success_rate(g, y):
    """Group success rate.

    This function computes the success rate for a given subgroup.

    Parameters
    ----------
    g : array-like
        subgroup vector (binary)
    y : array-like
        predictions vector (binary)

    Returns
    -------
    float
        group success rate

    """
    return y[g == 1].sum() / g.sum()  # success rate group_a


def statistical_parity(group_a, group_b, y_pred):
    """Statistical parity.

    This function computes the statistical parity (difference of success rates)\
    between group_a and group_b.

    Interpretation
    --------------
    A value of 0 is desired. Negative values are unfair towards group_a.\
    Positive values are unfair towards group_b. The range (-0.1,0.1)\
    is considered acceptable.

    Parameters
    ----------
    group_a : array-like
        Group membership vector (binary)
    group_b : array-like
        Group membership vector (binary)
    y_pred : array-like
        Predictions vector (binary)

    Returns
    -------
    float
        Statistical Parity

    Notes
    -----
    :math:`sr_a - sr_b`

    Examples
    --------
    >>> import numpy as np
    >>> from holisticai.bias.metrics import statistical_parity
    >>> group_a = np.array([1, 1, 1, 1, 0, 0, 0, 0, 0, 0])
    >>> group_b = np.array([0, 0, 0, 0, 1, 1, 1, 1, 1, 1])
    >>> y_pred = np.array([1, 1, 1, 0, 1, 1, 0, 0, 0, 0])
    >>> statistical_parity(group_a, group_b, y_pred)
    0.4166666666666667
    """
    # check and coerce
    group_a, group_b, y_pred, _ = _classification_checks(group_a, group_b, y_pred, y_true=None)

    # calculate sr_a and sr_b
    sr_a = _group_success_rate(group_a, y_pred)  # success rate group_a
    sr_b = _group_success_rate(group_b, y_pred)  # success rate group_b

    return sr_a - sr_b


def success_rate(group_a, group_b, y_pred):
    """Success Rate

    Calculates the raw success rates for each group.

    Parameters
    ----------
    group_a : array-like
        Group membership vector (binary)
    group_b : array-like
        Group membership vector (binary)
    y_pred : array-like
        Predictions vector (binary)

    Returns
    -------
    dict
        Dictionary with two keys, sr_a and sr_b (success rate for group a and b)
    """
    sr_a = _group_success_rate(group_a, y_pred)  # success rate group_a
    sr_b = _group_success_rate(group_b, y_pred)  # success rate group_b
    return {"sr_a": sr_a, "sr_b": sr_b}


def disparate_impact(group_a, group_b, y_pred):
    """Disparate Impact.

    This function computes the disparate impact (ratio of success rates)\
    between group_a and group_b class.

    Interpretation
    --------------
    A value of 1 is desired. Values below 1 are unfair towards group_a.\
    Values above 1 are unfair towards group_b. The range (0.8,1.2)\
    is considered acceptable.

    Parameters
    ----------
    group_a : array-like
        Group membership vector (binary)
    group_b : array-like
        Group membership vector (binary)
    y_pred : array-like
        Predictions vector (binary)

    Returns
    -------
    float
        Disparate Impact

    Notes
    -----
    :math:`sr_a/sr_b`

    References
    ----------
    .. [1] `M B Zafar, I Valera, M G Rodriguez, K P. Gummadi (2017).
            Fairness Constraints: Mechanisms for Fair Classification, MPI-SWS
            <https://arxiv.org/pdf/1507.05259.pdf>`

    Examples
    -------
    >>> import numpy as np
    >>> from holisticai.bias.metrics import disparate_impact
    >>> group_a = np.array([1, 1, 1, 1, 0, 0, 0, 0, 0, 0])
    >>> group_b = np.array([0, 0, 0, 0, 1, 1, 1, 1, 1, 1])
    >>> y_pred = np.array([1, 1, 1, 0, 1, 1, 0, 0, 0, 0])
    >>> disparate_impact(group_a, group_b, y_pred)
    2.25
    """
    # check and coerce
    group_a, group_b, y_pred, _ = _classification_checks(group_a, group_b, y_pred, y_true=None)

    # calculate sr_a and sr_b
    sr_a = _group_success_rate(group_a, y_pred)  # success rate group_a
    sr_b = _group_success_rate(group_b, y_pred)  # success rate group_b

    return sr_a / sr_b


def four_fifths(group_a, group_b, y_pred):
    """Four Fifths

    This function computes the four fifths rule (ratio of success rates)\
    between group_a and group_b. We return the minimum of the ratio taken both ways.

    Interpretation
    --------------
    A value of 1 is desired. Values below 1 are unfair. The range (0.8,1) is considered acceptable.

    Parameters
    ----------
    group_a : array-like
        Group membership vector (binary)
    group_b : array-like
        Group membership vector (binary)
    y_pred : array-like
        Predictions vector (binary)

    Returns
    -------
    float
        Four Fifths

    Examples
    --------
    >>> import numpy as np
    >>> from holisticai.bias.metrics import four_fifths
    >>> group_a = np.array([1, 1, 1, 1, 0, 0, 0, 0, 0, 0])
    >>> group_b = np.array([0, 0, 0, 0, 1, 1, 1, 1, 1, 1])
    >>> y_pred = np.array([1, 1, 1, 0, 1, 1, 0, 0, 0, 0])
    >>> four_fifths(group_a, group_b, y_pred)
    0.4444444444444444
    """
    # check and coerce
    group_a, group_b, y_pred, _ = _classification_checks(group_a, group_b, y_pred, y_true=None)

    # calculate sr_a and sr_b
    sr_a = _group_success_rate(group_a, y_pred)  # success rate group_a
    sr_b = _group_success_rate(group_b, y_pred)  # success rate group_b

    return min(sr_a / sr_b, sr_b / sr_a)


def cohen_d(group_a, group_b, y_pred):
    """Cohen D

    This function computes the Cohen D statistic (normalised statistical parity)\
    between group_a and group_b.

    Interpretation
    --------------
    A value of 0 is desired. Negative values are unfair towards group_a.
    Positive values are unfair towards group_b. Reference values: 0.2 is\
    considered a small effect size, 0.5 is considered medium, 0.8 is considered large.

    Parameters
    ----------
    group_a : array-like
        Group membership vector (binary)
    group_b : array-like
        Group membership vector (binary)
    y_pred : array-like
        Predictions vector (binary)

    Returns
    -------
    float
        Cohen D :

    Notes
    -----
    :math:`\frac{sr_a-sr_b}{\texttt{std_pool}}`

    Examples
    --------
    >>> import numpy as np
    >>> from holisticai.bias.metrics import cohen_d
    >>> group_a = np.array([1, 1, 1, 1, 0, 0, 0, 0, 0, 0])
    >>> group_b = np.array([0, 0, 0, 0, 1, 1, 1, 1, 1, 1])
    >>> y_pred = np.array([1, 1, 0, 0, 1, 1, 0, 1, 1, 1])
    >>> cohen_d(group_a, group_b, y_pred)
    -0.7844645405527363
    """
    # check and coerce
    group_a, group_b, y_pred, _ = _classification_checks(group_a, group_b, y_pred, y_true=None)

    # calculate sr_a and sr_b
    sr_a = _group_success_rate(group_a, y_pred)  # success rate group_a
    sr_b = _group_success_rate(group_b, y_pred)  # success rate group_b

    # calculate STD_a and STD_b
    std_b = np.sqrt(sr_b * (1 - sr_b))
    std_a = np.sqrt(sr_a * (1 - sr_a))

    # calculate n_a and n_b
    n_a = group_a.sum()
    n_b = group_b.sum()

    # calculate poolSTD
    std_pool = np.sqrt(((n_b - 1) * std_b**2 + (n_a - 1) * std_a**2) / (n_a + n_b - 2))

    return (sr_a - sr_b) / std_pool


def z_test_diff(group_a, group_b, y_pred):
    """Z Test (Difference)

    This function computes the Z-test statistic for the difference\
    in success rates. Also known as 2-SD Statistic.

    Interpretation
    --------------
    A value of 0 is desired. This test considers the data unfair if\
    the computed value is greater than 2 or smaller than -2, indicating\
    a statistically significant difference in success rates.

    Parameters
    ----------
    group_a : array-like
        Group membership vector (binary)
    group_b : array-like
        Group membership vector (binary)
    y_pred : array-like
        Predictions vector (binary)

    Returns
    -------
    float
        Z test (difference version)

    Examples
    --------
    >>> import numpy as np
    >>> from holisticai.bias.metrics import z_test_diff
    >>> group_a = np.array([1, 1, 1, 1, 0, 0, 0, 0, 0, 0])
    >>> group_b = np.array([0, 0, 0, 0, 1, 1, 1, 1, 1, 1])
    >>> y_pred = np.array([1, 1, 1, 0, 1, 1, 0, 0, 0, 0])
    >>> z_test_diff(group_a, group_b, y_pred)
    1.290994449

    References
    ----------
    .. [1] `Morris (2001).
           Sample size requirements for adverse impact analysis
            <https://www.semanticscholar.org/paper/Sample-Size-Required-for-Adverse-Impact-Analysis-Morris/877f7acd7c646a21f4947166a07f41664dcabe95>`

    """
    # check and coerce
    group_a, group_b, y_pred, _ = _classification_checks(group_a, group_b, y_pred, y_true=None)

    # calculate sr_a and sr_b
    sr_a = _group_success_rate(group_a, y_pred)  # success rate group_a
    sr_b = _group_success_rate(group_b, y_pred)  # success rate group_b
    n_a = group_a.sum()
    n_b = group_b.sum()
    sr_tot = (sr_a * n_a + sr_b * n_b) / (n_a + n_b)
    n_tot = n_a + n_b

    # calculate p_a
    p_a = n_a / n_tot

    return (sr_a - sr_b) / np.sqrt((sr_tot * (1 - sr_tot)) / (n_tot * p_a * (1 - p_a)))


def z_test_ratio(group_a, group_b, y_pred):
    """Z Test (Ratio)

    This function computes the Z-test statistic for the ratio\
    in success rates. Also known as 2-SD Statistic.

    Interpretation
    --------------
    A value of 0 is desired. This test considers the data unfair if\
    the computed value is greater than 2 or smaller than -2, indicating\
    a statistically significant ratio in success rates.

    Parameters
    ----------
    group_a : array-like
        Group membership vector (binary)
    group_b : array-like
        Group membership vector (binary)
    y_pred : array-like
        Predictions vector (binary)

    Returns
    -------
    float
        Z-test (ratio version)

    Examples
    --------
    >>> import numpy as np
    >>> from holisticai.bias.metrics import z_test_ratio
    >>> group_a = np.array([1, 1, 1, 1, 0, 0, 0, 0, 0, 0])
    >>> group_b = np.array([0, 0, 0, 0, 1, 1, 1, 1, 1, 1])
    >>> y_pred = np.array([1, 1, 1, 0, 1, 1, 0, 0, 0, 0])
    >>> z_test_ratio(group_a, group_b, y_pred)
    1.256287689

    References
    ----------
    .. [1] `Morris (2001).
           Sample size requirements for adverse impact analysis
            <https://www.semanticscholar.org/paper/Sample-Size-Required-for-Adverse-Impact-Analysis-Morris/877f7acd7c646a21f4947166a07f41664dcabe95>`
    """
    # check and coerce
    group_a, group_b, y_pred, _ = _classification_checks(group_a, group_b, y_pred, y_true=None)

    # calculate sr_a and sr_b
    sr_a = _group_success_rate(group_a, y_pred)  # success rate group_a
    sr_b = _group_success_rate(group_b, y_pred)  # success rate group_b
    n_a = group_a.sum()
    n_b = group_b.sum()
    sr_tot = (sr_a * n_a + sr_b * n_b) / (n_a + n_b)
    n_tot = n_a + n_b

    # calculate p_a
    p_a = n_a / n_tot

    return (np.log(sr_a / sr_b)) / np.sqrt((1 - sr_tot) / (sr_tot * n_tot * p_a * (1 - p_a)))


def _correlation_diff(group_a, group_b, y_pred, y_true):
    """Correlation difference

    This function computes the difference in correlation between predicted\
    and true labels for group_a and group_b.

    Interpretation
    --------------
    A value of 0 is desired. This metric ranges between -2 and 2, with\
    negative values indicating bias against group_a, and positive values indicating bias against group_b.

    Parameters
    ----------
    group_a : array-like
        Group membership vector (binary)
    group_b : array-like
        Group membership vector (binary)
    y_pred : array-like
        Predictions vector (binary)
    y_true : array-like
        Target vector (binary)

    Returns
    -------
    float
        Correlation Difference

    Notes
    -----
    :math:`CV_a - CV_b`

    Examples
    --------
    >>> import numpy as np
    >>> from holisticai.bias.metrics import correlation_diff
    >>> group_a = np.array([1, 1, 1, 1, 0, 0, 0, 0, 0, 0])
    >>> group_b = np.array([0, 0, 0, 0, 1, 1, 1, 1, 1, 1])
    >>> y_pred = np.array([1, 1, 0, 0, 1, 1, 0, 1, 1, 1])
    >>> y_true = np.array([1, 1, 0, 0, 1, 0, 1, 0, 0, 1])
    >>> correlation_diff(group_a, group_b, y_pred, y_true)
    1.4472135954999579
    """
    # check and coerce
    group_a, group_b, y_pred, y_true = _classification_checks(group_a, group_b, y_pred, y_true)

    # Calculate Pearson correlations
    cv_a = np.corrcoef(y_pred[group_a == 1], y_true[group_a == 1])[1, 0]
    cv_b = np.corrcoef(y_pred[group_b == 1], y_true[group_b == 1])[1, 0]

    return cv_a - cv_b


def equal_opportunity_diff(group_a, group_b, y_pred, y_true):
    """Equality of opportunity difference

    This function computes the difference in true positive\
    rates for group_a and group_b.

    Interpretation
    --------------
    A value of 0 is desired. This metric ranges between -1 and 1,\
    with negative values indicating bias against group_a, and\
    positive values indicating bias against group_b.

    Parameters
    ----------
    group_a : array-like
        Group membership vector (binary)
    group_b : array-like
        Group membership vector (binary)
    y_pred : array-like
        Predictions vector (binary)
    y_true : array-like
        Target vector (binary)

    Returns
    -------
    float
        Equal opportunity difference

    Notes
    -----
    :math:`tpr_a - tpr_b`

    Examples
    --------
    >>> import numpy as np
    >>> from holisticai.bias.metrics import equal_opportunity_diff
    >>> group_a = np.array([1, 1, 1, 1, 0, 0, 0, 0, 0, 0])
    >>> group_b = np.array([0, 0, 0, 0, 1, 1, 1, 1, 1, 1])
    >>> y_pred = np.array([1, 1, 0, 0, 1, 1, 0, 1, 1, 1])
    >>> y_true = np.array([1, 1, 0, 0, 1, 0, 1, 0, 0, 1])
    >>> equal_opportunity_diff(group_a, group_b, y_pred, y_true)
    0.33333333333333337
    """
    # check and coerce
    group_a, group_b, y_pred, y_true = _classification_checks(group_a, group_b, y_pred, y_true)

    # Calculate true positive rates
    tpr_a = confusion_matrix(y_true[group_a == 1], y_pred[group_a == 1], normalize="true")[1, 1]
    tpr_b = confusion_matrix(y_true[group_b == 1], y_pred[group_b == 1], normalize="true")[1, 1]

    return tpr_a - tpr_b


def false_positive_rate_diff(group_a, group_b, y_pred, y_true):
    """False positive rate difference

    This function computes the difference in false positive\
    rates between group_a and group_b.

    Interpretation
    --------------
    A value of 0 is desired. This metric ranges between -1 and 1,\
    with negative values indicating bias against group_a, and\
    positive values indicating bias against group_b.

    Parameters
    ----------
    group_a : array-like
        Group membership vector (binary)
    group_b : array-like
        Group membership vector (binary)
    y_pred : array-like
        Predictions vector (binary)
    y_true : array-like
        Target vector (binary)

    Returns
    -------
    float
        FPR_diff

    Notes
    -----
    :math:`fpr_a - fpr_b`

    Examples
    --------
    >>> import numpy as np
    >>> from holisticai.bias.metrics import false_positive_diff
    >>> group_a = np.array([1, 1, 1, 1, 0, 0, 0, 0, 0, 0])
    >>> group_b = np.array([0, 0, 0, 0, 1, 1, 1, 1, 1, 1])
    >>> y_pred = np.array([1, 1, 0, 0, 1, 1, 0, 1, 1, 1])
    >>> y_true = np.array([1, 1, 0, 0, 1, 0, 1, 0, 0, 1])
    >>> false_positive_diff(group_a, group_b, y_pred, y_true)
    -1.0
    """
    # check and coerce
    group_a, group_b, y_pred, y_true = _classification_checks(group_a, group_b, y_pred, y_true)

    # Calculate false positive rates
    fpr_a = confusion_matrix(y_true[group_a == 1], y_pred[group_a == 1], normalize="true")[0, 1]
    fpr_b = confusion_matrix(y_true[group_b == 1], y_pred[group_b == 1], normalize="true")[0, 1]

    return fpr_a - fpr_b


def false_negative_rate_diff(group_a, group_b, y_pred, y_true):
    """False negative Rate difference

    This function computes the difference in false negative\
    rates for group_a and group_b.

    Interpretation
    ----------
    A value of 0 is desired. This metric ranges between -1 and 1,\
    with negative values indicating bias against group_b, and\
    positive values indicating bias against group_a.

    Parameters
    ----------
    group_a : array-like
        Group membership vector (binary)
    group_b : array-like
        Group membership vector (binary)
    y_pred : array-like
        Predictions vector (binary)
    y_true : array-like
        Target vector (binary)

    Returns
    -------
    float
        False Negative Rate difference

    Notes
    -----
    :math:`fnr_a - fnr_b`

    Examples
    -------
    >>> import numpy as np
    >>> from holisticai.bias.metrics import fnr_diff
    >>> group_a = np.array([1, 1, 1, 1, 0, 0, 0, 0, 0, 0])
    >>> group_b = np.array([0, 0, 0, 0, 1, 1, 1, 1, 1, 1])
    >>> y_pred = np.array([1, 1, 0, 0, 1, 1, 0, 1, 1, 1])
    >>> y_true = np.array([1, 1, 0, 0, 1, 0, 1, 0, 0, 1])
    >>> fnr_diff(group_a, group_b, y_pred, y_true)
    -0.3333333333333333
    """
    # check and coerce
    group_a, group_b, y_pred, y_true = _classification_checks(group_a, group_b, y_pred, y_true)

    # Calculate false negative rates
    fnr_a = confusion_matrix(y_true[group_a == 1], y_pred[group_a == 1], normalize="true")[1, 0]
    fnr_b = confusion_matrix(y_true[group_b == 1], y_pred[group_b == 1], normalize="true")[1, 0]

    return fnr_a - fnr_b


def true_negative_rate_diff(group_a, group_b, y_pred, y_true):
    """True negative Rate difference

    This function computes the difference in true negative\
    rates for group_a and group_b.

    Interpretation
    ----------
    A value of 0 is desired. This metric ranges between -1 and 1,\
    with negative values indicating bias against group_a, and\
    positive values indicating bias against group_b.

    Parameters
    ----------
    group_a : array-like
        Group membership vector (binary)
    group_b : array-like
        Group membership vector (binary)
    y_pred : array-like
        Predictions vector (binary)
    y_true : array-like
        Target vector (binary)

    Returns
    -------
    float
        True Negative Rate difference

    Notes
    -----
    :math:`tnr_a - tnr_b`

    Examples
    -------
    >>> import numpy as np
    >>> from holisticai.bias.metrics import tnr_diff
    >>> group_a = np.array([1, 1, 1, 1, 0, 0, 0, 0, 0, 0])
    >>> group_b = np.array([0, 0, 0, 0, 1, 1, 1, 1, 1, 1])
    >>> y_pred = np.array([1, 1, 0, 0, 1, 1, 0, 1, 1, 1])
    >>> y_true = np.array([1, 1, 0, 0, 1, 0, 1, 0, 0, 1])
    >>> tnr_diff(group_a, group_b, y_pred, y_true)
    1
    """
    # check and coerce
    group_a, group_b, y_pred, y_true = _classification_checks(group_a, group_b, y_pred, y_true)

    # Calculate false negative rates
    tnr_a = confusion_matrix(y_true[group_a == 1], y_pred[group_a == 1], normalize="true")[0, 0]
    tnr_b = confusion_matrix(y_true[group_b == 1], y_pred[group_b == 1], normalize="true")[0, 0]

    return tnr_a - tnr_b


def average_odds_diff(group_a, group_b, y_pred, y_true):
    """Average Odds Difference

    This function computes the difference in average odds\
    between group_a and group_b.

    Interpretation
    --------------
    A value of 0 is desired. This metric ranges between -1 and 1,\
    with negative values indicating bias against group_a,\
    and positive values indicating bias against group_b.\
    The range (-0.1,0.1) is considered acceptable.

    Parameters
    ----------
    group_a : array-like
        Group membership vector (binary)
    group_b : array-like
        Group membership vector (binary)
    y_pred : array-like
        Predictions vector (binary)
    y_true : array-like
        Target vector (binary)

    Returns
    -------
    float
        AOD score

    Notes
    -----
    :math:`0.5 * (fpr_a-fpr_b + tpr_a-tpr_b)`

    Examples
    -------
    >>> import numpy as np
    >>> from holisticai.bias.metrics import average_odds_diff
    >>> group_a = np.array([1, 1, 1, 1, 0, 0, 0, 0, 0, 0])
    >>> group_b = np.array([0, 0, 0, 0, 1, 1, 1, 1, 1, 1])
    >>> y_pred = np.array([1, 1, 0, 0, 1, 1, 0, 1, 1, 1])
    >>> y_true = np.array([1, 1, 0, 0, 1, 0, 1, 0, 0, 1])
    >>> average_odds_diff(group_a, group_b, y_pred, y_true)
    -0.3333333333333333
    """
    # check and coerce
    group_a, group_b, y_pred, y_true = _classification_checks(group_a, group_b, y_pred, y_true)

    # Compute AOD
    return 0.5 * (
        equal_opportunity_diff(group_a, group_b, y_pred, y_true)
        + false_positive_rate_diff(group_a, group_b, y_pred, y_true)
    )


def accuracy_diff(group_a, group_b, y_score, y_true):
    """Accuracy Difference

    This function computes the difference in accuracy\
    of predictions for group_a and group_b

    Interpretation
    --------------
    A value of 0 is desired. This metric ranges between -1 and 1,\
    with negative values indicating bias against group_a,\
    and positive values indicating bias against group_b.

    Parameters
    ----------
    group_a : array-like
        Group membership vector (binary)
    group_b : array-like
        Group membership vector (binary)
    y_score : numpy array
        Probability estimates (regression)
    y_true : array-like
        Target vector (binary)

    Returns
    -------
    float
        acc_diff : acc_a - acc_b

    Examples
    -------
    >>> import numpy as np
    >>> from holisticai.bias.metrics import accuracy_diff
    >>> group_a = np.array([1, 1, 1, 1, 0, 0, 0, 0, 0, 0])
    >>> group_b = np.array([0, 0, 0, 0, 1, 1, 1, 1, 1, 1])
    >>> y_pred = np.array([1, 1, 0, 0, 1, 1, 0, 1, 1, 1])
    >>> y_true = np.array([1, 1, 0, 0, 1, 0, 1, 0, 0, 1])
    >>> accuracy_diff(group_a, group_b, y_pred, y_true)
    0.6666666666666667
    """
    # check and coerce
    group_a, group_b, y_score, y_true, _ = _regression_checks(group_a, group_b, y_score, y_true, None)
    _check_binary(y_true, "y_true")

    # split data by groups
    y_true_a = y_true[group_a == 1]
    y_score_a = y_score[group_a == 1]
    y_true_b = y_true[group_b == 1]
    y_score_b = y_score[group_b == 1]

    # compute abroca
    return accuracy_score(y_true_a, y_score_a) - accuracy_score(y_true_b, y_score_b)


def abroca(group_a, group_b, y_score, y_true):
    """ABROCA (area between roc curves)

    This function computes the area between the roc curve\
    of group_a and the roc curve of group_b

    Interpretation
    --------------
    A value of 0 is desired. This metric ranges between -1 and 1,\
    with negative values indicating bias against group_a,\
    and positive values indicating bias against group_b.

    Parameters
    ----------
    group_a : array-like
        Group membership vector (binary)
    group_b : array-like
        Group membership vector (binary)
    y_score : array-like
        Probability estimates (regression)
    y_true : array-like
        Target vector (binary)

    Returns
    -------
    float
        ABROCA : roc_auc_a - roc_auc_b

    Examples
    --------
    >>> import numpy as np
    >>> from holisticai.bias.metrics import abroca
    >>> group_a = np.array([1] * 50 + [0] * 50)
    >>> group_b = np.array([0] * 50 + [1] * 50)
    >>> y_score = np.concatenate((np.linspace(0, 1, 50), np.linspace(0, 1, 50) ** 2))
    >>> y_true = y_score + np.random.random(y_score.shape) > 0.5
    >>> abroca(group_a, group_b, y_score, y_true)
    0.11806478405315601
    """
    # check and coerce
    group_a, group_b, y_score, y_true, _ = _regression_checks(group_a, group_b, y_score, y_true, None)
    _check_binary(y_true, "y_true")

    # split data by groups
    y_true_a = y_true[group_a == 1]
    y_score_a = y_score[group_a == 1]
    y_true_b = y_true[group_b == 1]
    y_score_b = y_score[group_b == 1]

    # compute abroca
    return roc_auc_score(y_true_a, y_score_a) - roc_auc_score(y_true_b, y_score_b)


def classification_bias_metrics(
    group_a=None, group_b=None, y_pred=None, y_true=None, y_score=None, X=None, metric_type="group", **kargs
):
    """Classification bias metrics batch computation

    This function computes all the relevant classification bias metrics,\
    and displays them as a pandas dataframe. It also includes a fair reference\
    value for comparison.

    Parameters
    ----------
    group_a : array-like
        Group membership vector (binary)
    group_b : array-like
        Group membership vector (binary)
    y_pred : array-like
        Predictions vector (binary)
    y_true : array-like, optional
        Target vector (binary)
    y_score : array-like, optional
        Probability estimates (regression)
    X : array-like, optional
        Feature matrix
    metric_type : str, optional
        Specifies which metrics we compute 'group', 'individual', 'equal_outcome' , 'equal_opportunity'
    **kargs : dict
        Additional keyword arguments for individual metrics

    Returns
    -------
    pandas DataFrame
        Metrics | Values | Reference
    """

    individual_metrics_format_1 = {
        "Theil Index": theil_index,
        "Generalized Entropy Index": generalized_entropy_index,
        "Coefficient of Variation": coefficient_of_variation,
    }

    individual_metrics_format_2 = {
        "Consistency Score": consistency_score,
    }

    equal_outcome_metrics = {
        "Statistical Parity": statistical_parity,
        "Disparate Impact": disparate_impact,
        "Four Fifths Rule": four_fifths,
        "Cohen D": cohen_d,
        "2SD Rule": z_test_diff,
    }

    equal_opportunity_metrics = {
        "Equality of Opportunity Difference": equal_opportunity_diff,
        "False Positive Rate Difference": false_positive_rate_diff,
        "Average Odds Difference": average_odds_diff,
        "Accuracy Difference": accuracy_diff,
    }

    soft_metrics = {
        "ABROCA": abroca,
    }

    ref_vals = {
        "Statistical Parity": 0,
        "Disparate Impact": 1,
        "Four Fifths Rule": 1,
        "Cohen D": 0,
        "Equality of Opportunity Difference": 0,
        "False Positive Rate Difference": 0,
        "Average Odds Difference": 0,
        "Accuracy Difference": 0,
        "ABROCA": 0,
        "2SD Rule": 0,
        "Theil Index": 0,
        "Generalized Entropy Index": 0,
        "Consistency Score": 1,
        "Coefficient of Variation": 0,
    }

    has_group_parameters = all((p is not None) for p in [group_a, group_b, y_pred])

    if has_group_parameters:
        out_metrics = [[pf, fn(group_a, group_b, y_pred), ref_vals[pf]] for pf, fn in equal_outcome_metrics.items()]
        opp_metrics = []
        if y_true is not None:
            opp_metrics += [
                [pf, fn(group_a, group_b, y_pred, y_true), ref_vals[pf]] for pf, fn in equal_opportunity_metrics.items()
            ]
        if y_score is not None:
            opp_metrics += [
                [pf, fn(group_a, group_b, y_score, y_true), ref_vals[pf]] for pf, fn in soft_metrics.items()
            ]

    if metric_type == "individual":
        from collections import defaultdict

        metric_kargs = defaultdict(dict)
        for k, value in kargs.items():
            metric, arg = k.split("__")
            metric_kargs[metric][arg] = value

        indv_metrics = []
        if y_pred is not None:
            if y_true is not None:
                indv_metrics += [
                    [pf, fn(y_pred, y_true, **metric_kargs[fn.__name__]), ref_vals[pf]]
                    for pf, fn in individual_metrics_format_1.items()
                ]

            if X is not None:
                indv_metrics += [
                    [pf, fn(X, y_pred, **metric_kargs[fn.__name__]), ref_vals[pf]]
                    for pf, fn in individual_metrics_format_2.items()
                ]

    if metric_type in ["group", "both"]:
        if metric_type == "both":
            # TODO: remove both for next version
            warnings.warn(  # noqa: B028
                "`both` option will be depreciated in the next versions, use group",
                DeprecationWarning,
            )

        metrics = out_metrics + opp_metrics
        return pd.DataFrame(metrics, columns=["Metric", "Value", "Reference"]).set_index("Metric")

    if metric_type == "equal_outcome":
        return pd.DataFrame(out_metrics, columns=["Metric", "Value", "Reference"]).set_index("Metric")

    if metric_type == "equal_opportunity":
        return pd.DataFrame(opp_metrics, columns=["Metric", "Value", "Reference"]).set_index("Metric")

    if metric_type == "individual":
        return pd.DataFrame(indv_metrics, columns=["Metric", "Value", "Reference"]).set_index("Metric")

    msg = "metric_type is not one of : group, individual, equal_outcome, equal_opportunity"
    raise ValueError(msg)


### Individual Metrics


def benefit_function(y_pred, y_true):
    """Benefit function

    This function computes the benefit function\
    used in the generalized entropy index.

    Parameters
    ----------
    y_true : array-like
        True target values
    y_pred : array-like
        Predicted target values

    Returns
    -------
    np.ndarray
        Benefit function
    """
    return y_pred - y_true + 1


def theil_index(y_pred: np.ndarray, y_true: np.ndarray) -> float:
    """The Theil index

    The Theil index is a measure of inequality that is commonly used in economics.\
    It is used to measure the inequality of a distribution, such as the distribution\
    of income or wealth. The Theil index is a special case of general entropy indices\
    that allows to observe inequalities at group level and individual level.

    Interpretation
    --------------
    A high value implies high inequality while a value of 0 indicates perfect equality.

    Parameters
    ----------
    y_true : array-like of shape (n_samples,)
        The true target values.
    y_pred : array-like of shape (n_samples,)
        The predicted target values.

    Returns
    -------
    float
        The Theil index of the input array.

    Examples
    -------
    >>> import numpy as np
    >>> from holisticai.bias.metrics import theil_index
    >>> y_true = np.array([1, 2, 3, 4, 5])
    >>> y_pred = np.array([1, 2, 3, 4, 5])
    >>> theil_index(y_pred, y_true)
    0.0
    """
    return generalized_entropy_index(y_pred, y_true, alpha=1)


def generalized_entropy_index(y_pred: np.ndarray, y_true: np.ndarray, alpha: int = 2) -> float:
    """Generalized entropy index

    Generalized entropy index is a measure of inequality that is proposed as a unified\
    individual and group fairness measure [1]. It is used to measure the extent of inequality\
    in the distribution of a variable. The measure is based on the idea that inequality can\
    be measured by the extent to which the distribution of a variable deviates from a hypothetical\
    distribution in which everyone has an equal share.

    Interpretation
    --------------
    A high value implies high inequality while a value of 0 indicates perfect equality.

    Parameters
    ----------
    y_true : array-like of shape (n_samples,)
        The true target values.
    y_pred : array-like of shape (n_samples,)
        The predicted target values.
    alpha : int, optional, default=2
        Parameter that regulates the weight given to distances between values at different parts\
        of the distribution. Default=2.

    Returns
    -------
    float
        The generalized entropy index of the input arrays.

    References
    ----------
    .. [1] Speicher, T., Heidari, H., Grgic-Hlaca, N., Gummadi, K. P., & Weller, A. (2018, April).\
        A Unified Approach to Quantifying Algorithmic Unfairness: Measuring Individual & Group Unfairness\
        via Inequality Indices. In Proceedings of the 24th ACM SIGKDD International Conference on Knowledge\
        Discovery & Data Mining (pp. 2239-2248).

    Examples
    --------
    >>> import numpy as np
    >>> from holisticai.bias.metrics import generalized_entropy_index
    >>> y_true = np.array([1, 2, 3, 4, 5])
    >>> y_pred = np.array([1, 2, 3, 4, 5])
    >>> generalized_entropy_index(y_true, y_pred, alpha=2)
    0.0
    """
    y_pred = _array_like_to_numpy(y_pred)
    y_true = _array_like_to_numpy(y_true)
    _check_binary(y_true)

    b = benefit_function(y_pred, y_true)
    u = np.mean(b)
    bu = b / u

    if alpha == 1:
        # rewrite equation to allow 0 b values
        return np.mean(np.log(bu**b) / u)

    if alpha == 0:
        return -np.mean(np.log(bu) / u)

    return (1 / (alpha * (alpha - 1))) * np.mean(bu**alpha - 1)


def coefficient_of_variation(y_pred: np.ndarray, y_true: np.ndarray) -> float:
    """Coefficient of variation

    Calculates the coefficient of variation, which is a special case that calculates two times the square\
    root of the generalized entropy index with alpha=2.

    Interpretation
    --------------
    A high value implies high inequality while a value of 0 indicates perfect equality.

    Parameters
    ----------
    y_true : array-like
        True target values.
    y_pred : array-like
        Predicted target values.

    Returns
    --------
    float
        The coefficient of variation.
    """
    y_pred = _array_like_to_numpy(y_pred)
    y_true = _array_like_to_numpy(y_true)
    _check_binary(y_true)
    return 2 * np.sqrt(generalized_entropy_index(y_pred, y_true, alpha=2))


def consistency_score(X: np.ndarray, y_pred: np.ndarray, n_neighbors: int = 5) -> float:
    """Consistency score

    Compares a model's classification prediction of a given data item X to its k-nearest neighbors.

    Parameters
    ----------
    X : array-like
        The input features.
    y_pred : array-like
        The predicted labels.
    n_neighbors : int, optional
        Number of neighbors for the knn computation. Default is 5.

    Returns
    -------
    float
        The consistency score.
    """
    X = _matrix_like_to_numpy(X)
    y_pred = _array_like_to_numpy(y_pred)
    _check_binary(y_pred)

    # learn a KNN on the features
    nbrs = NearestNeighbors(n_neighbors=n_neighbors, algorithm="ball_tree").fit(X)
    _, indices = nbrs.kneighbors(X)

    # compute consistency score
    return 1 - np.mean([np.abs(yi - np.mean(y_pred[ids])) for yi, ids in zip(y_pred, indices)])



from holisticai.security.metrics import shapr_score
from holisticai.datasets import load_dataset
from sklearn.tree import DecisionTreeClassifier
from sklearn.preprocessing import StandardScaler, OneHotEncoder
from sklearn.compose import ColumnTransformer
from sklearn.pipeline import Pipeline
from tests.security.utils import categorical_dataset

def create_preprocessor(X):
    categorical_features = X.select_dtypes(include=['category']).columns
    numerical_fatures = X.select_dtypes(exclude=['category']).columns

    # Create transformers for numerical and categorical features
    numeric_transformer = Pipeline(steps=[('scaler', StandardScaler())])
    categorical_transformer = Pipeline(steps=[('onehot', OneHotEncoder(handle_unknown='ignore'))])

    # Combine transformers into a preprocessor using ColumnTransformer
    return ColumnTransformer(
        transformers=[
            ('num', numeric_transformer, numerical_fatures),
            ('cat', categorical_transformer, categorical_features)
    ])

def test_shapr(categorical_dataset):
    train = categorical_dataset['train']
    test = categorical_dataset['test']

    preprocessor = create_preprocessor(train['X'])
    model = Pipeline(steps=[('preprocessor', preprocessor),
                            ('classifier', DecisionTreeClassifier())])

    model.fit(train['X'], train['y'])

    pred_train = model.predict(train['X'])
    pred_test = model.predict(test['X'])
    v = shapr_score(train['y'], test['y'], pred_train, pred_test) 
    print(v)



import numpy as np
import pandas as pd


def quantil_classify(q1, q2, q3, labels, x):
    if x <= q1:
        return labels[0]
    if q1 < x <= q2:
        return labels[1]
    if q2 < x <= q3:
        return labels[2]
    return labels[3]


def group_samples_by_learning_task(y: pd.Series, learning_task="regression", return_group_mask=False):
    if learning_task in ["binary_classification", "multi_classification"]:
        y_group = y.astype(str)

    elif learning_task in ["regression"]:
        labels = ["Q0-Q1", "Q1-Q2", "Q2-Q3", "Q3-Q4"]
        labels_values = [0.0, 0.25, 0.5, 0.75, 1.0]
        v = np.array(y.quantile(labels_values)).squeeze()
        y_group = y.map(lambda x: quantil_classify(v[1], v[2], v[3], labels, x))

    else:
        raise ValueError(f"Learning task {learning_task} not supported")

    if return_group_mask:
        return y_group
    return y_group.groupby(y_group).apply(lambda x: x.index.tolist()).to_dict()



import pytest
from holisticai.datasets import load_dataset
import numpy as np
from sklearn.linear_model import LogisticRegression
from numpy.random import RandomState

@pytest.fixture
def input_data():
    dataset = load_dataset("adult")
    dataset = dataset.sample(n=1000, random_state=42)
    dataset = dataset.train_test_split(test_size=0.3)
    train = dataset['train']
    test = dataset['test']

    model = LogisticRegression(random_state=42)
    model.fit(train['X'], train['y'])

    return test, model

def test_get_partial_dependence_results_binary_classification(input_data):
    test,model = input_data
    from holisticai.utils import RegressionProxy
    from holisticai.utils.feature_importances import compute_permutation_feature_importance
    from holisticai.utils.inspection import compute_partial_dependence
    
    proxy = RegressionProxy(predict=model.predict)
    importances  = compute_permutation_feature_importance(proxy=proxy, X=test['X'], y=test['y'])
    ranked_importances = importances.top_alpha(0.8)
    partial_dependencies = compute_partial_dependence(test['X'], features=ranked_importances.feature_names, proxy=proxy)
    np.isclose(partial_dependencies.values[0][0]['average'][0][0], 0.03)


from typing import Literal, Union, overload

@overload
def load_dataset(
    dataset_name: Literal["adult"],
    protected_attribute: Union[Literal["race", "sex"], None] = None,
    preprocessed: bool = True,
): ...
@overload
def load_dataset(
    dataset_name: Literal["law_school"],
    protected_attribute: Union[Literal["race", "gender"], None] = None,
    preprocessed: bool = True,
): ...
@overload
def load_dataset(
    dataset_name: Literal["student_multiclass"],
    protected_attribute: Union[Literal["sex", "address"], None] = None,
    preprocessed: bool = True,
): ...
@overload
def load_dataset(
    dataset_name: Literal["student"],
    target: Union[Literal["G1", "G2", "G3"], None] = None,
    preprocessed: bool = True,
    protected_attribute: Union[Literal["sex", "address"], None] = None,
): ...
@overload
def load_dataset(
    dataset_name: Literal["lastfm"],
): ...
@overload
def load_dataset(
    dataset_name: Literal["us_crime"],
    preprocessed: bool = True,
    protected_attribute: Union[Literal["race"], None] = None,
): ...
@overload
def load_dataset(
    dataset_name: Literal["us_crime_multiclass"],
    preprocessed: bool = True,
    protected_attribute: Union[Literal["race"], None] = None,
): ...
@overload
def load_dataset(
    dataset_name: Literal["clinical_records"], protected_attribute: Union[Literal["sex"], None] = None
): ...



"""
The :mod:holisticai.privacy.metrics module includes attacks and privacy metrics
"""

from __future__ import annotations

from typing import Union

import pandas as pd
from holisticai.security.metrics._anonymization import k_anonymity, l_diversity
from holisticai.security.metrics._attribute_attack import AttributeAttackScore, attribute_attack_score
from holisticai.security.metrics._data_minimization import (
    DataMinimizationAccuracyRatio,
    DataMinimizationMSERatio,
    data_minimization_score,
)
from holisticai.security.metrics._shapr import ShaprScore, shapr_score


def classification_privacy_metrics(
    x_train: pd.DataFrame,
    x_test: pd.DataFrame,
    y_train: pd.Series,
    y_test: pd.Series,
    y_pred_train: pd.Series,
    y_pred_test: pd.Series,
    y_pred_test_dm: dict[str, : pd.Series],
    attribute_attack: Union[str, list[str]],
):
    shapr_score = ShaprScore()
    dm_accuracy_ratio = DataMinimizationAccuracyRatio()
    attr_attack_score = AttributeAttackScore()

    results = []
    value = shapr_score(y_train, y_test, y_pred_train, y_pred_test)
    results.append({"metric": shapr_score.name, "value": value, "reference": shapr_score.reference})

    value = dm_accuracy_ratio(y_test, y_pred_test, y_pred_test_dm)
    results.append({"metric": dm_accuracy_ratio.name, "value": value, "reference": dm_accuracy_ratio.reference})

    value = attr_attack_score(x_train, x_test, y_train, y_test, attribute_attack=attribute_attack)
    results.append({"metric": attr_attack_score.name, "value": value, "reference": attr_attack_score.reference})

    return pd.DataFrame(results)


def regression_privacy_metrics(
    x_train: pd.DataFrame,
    x_test: pd.DataFrame,
    y_train: pd.Series,
    y_test: pd.Series,
    y_pred_test: pd.Series,
    y_pred_test_dm: dict[str, : pd.Series],
    attribute_attack: Union[str, list[str]],
):
    attr_attack_score = AttributeAttackScore()
    dm_mse_ratio = DataMinimizationMSERatio()

    results = []
    value = dm_mse_ratio(y_test, y_pred_test, y_pred_test_dm)
    results.append({"metric": dm_mse_ratio.name, "value": value, "reference": dm_mse_ratio.reference})

    value = attr_attack_score(
        x_train, x_test, y_train, y_test, learning_task="regression", attribute_attack=attribute_attack
    )
    results.append({"metric": attr_attack_score.name, "value": value, "reference": attr_attack_score.reference})

    return pd.DataFrame(results)


__all__ = [
    "k_anonymity",
    "l_diversity",
    "classification_privacy_metrics",
    "shapr_score",
    "ShaprScore",
    "DataMinimizationAccuracyRatio",
    "data_minimization_score",
    "attribute_attack_score",
    "AttributeAttackScore",
]



import matplotlib.pyplot as plt
import numpy as np
from holisticai.explainability.metrics.global_importance._xai_ease_score import XAIEaseAnnotator
from holisticai.utils import Importances, PartialDependence


def plot_partial_dependence(
    partial_dependence: PartialDependence,
    ranked_feature_importance: Importances,
    subplots=(1, 1),
    figsize=None,
    class_idx=0,
):
    """
    Plots the partial dependence of features on the predicted target.

    Parameters
    ----------
    partial_dependence: PartialDependence
            The partial dependence object containing the computed partial dependence values.
    ranked_feature_importance: RankedFeatureImportance
        The ranked feature importance object containing the feature names and their importance scores.
    subplots: (tuple, optional)
        The shape of the subplots grid. Defaults to (1, 1).
    figsize: (tuple, optional)
        The size of the figure. Defaults to None.

    Returns
    -------
        fig: The matplotlib figure object containing the plot.

    Example
    -------
    >>> partial_dependence = PartialDependence(values=[...])
    >>> ranked_feature_importance = Importances(values=[...], feature_names=[...])
    >>> plot_partial_dependence(partial_dependence, ranked_feature_importance)

    The plot should look like this:

    .. image:: /_static/images/xai_plot_partial_dependence.png
        :alt: Plot Partial Dependence
    """
    partial_dependence_values = partial_dependence.values[class_idx]
    _, axs = plt.subplots(*subplots, figsize=figsize)
    axs = [axs] if isinstance(axs, plt.Axes) else axs.flatten()
    n_plots = min(len(axs), len(partial_dependence_values))
    annotator = XAIEaseAnnotator()
    for feature_index in range(n_plots):
        ax = axs[feature_index]
        individuals = partial_dependence_values[feature_index]["individual"][0]
        average = partial_dependence_values[feature_index]["average"][0]
        x = partial_dependence_values[feature_index]["grid_values"][0]
        level = annotator.compute_xai_ease_score_data(partial_dependence_values, ranked_feature_importance).set_index(
            "feature"
        )["scores"]
        feature_name = ranked_feature_importance.feature_names[feature_index]
        feature_value = ranked_feature_importance[feature_index]

        curve_media = np.mean(individuals, axis=0)
        curve_std = np.std(individuals, axis=0)
        curve_lower = curve_media - curve_std
        curve_upper = curve_media + curve_std

        ax.plot(x, average, color="blue", label=level.loc[feature_name])
        # for curve in individuals:
        #    ax.plot(x, curve, alpha=0.05, color="skyblue")
        ax.fill_between(x, curve_lower, curve_upper, color="skyblue", alpha=0.2)

        ymin = individuals.min()
        ymax = individuals.max()
        ax.set_ylim(ymin, ymax)

        xmin = x.min()
        xmax = x.max()
        ax.set_xlim(xmin, xmax)
        ax.set_xlabel("Feature Value")
        ax.set_ylabel("Predicted Target")

        ax.set_title(f"{feature_name} ({feature_value:.3f})")
        ax.grid(True)
        ax.legend()
    plt.tight_layout()



from holisticai.datasets import load_dataset
from sklearn.naive_bayes import GaussianNB
from holisticai.explainability.metrics import classification_explainability_metrics
from holisticai.explainability.metrics.global_importance import alpha_score, rank_alignment, position_parity, xai_ease_score
import numpy as np
import pytest

@pytest.fixture
def input_data():
    dataset = load_dataset('adult').select(list(range(1000)))
    dataset = dataset.train_test_split(test_size=0.2, random_state=42)
    train = dataset['test']
    test = dataset['test']
    model = GaussianNB()
    model.fit(train['X'], train['y'])
    return model, test

def get_classification_features(model, test, strategy):
    from holisticai.utils import BinaryClassificationProxy
    from holisticai.utils.inspection import compute_partial_dependence

    if strategy=='permutation':
        from holisticai.utils.feature_importances import compute_permutation_feature_importance as compute_feature_importance
    elif strategy=='surrogate':
        from holisticai.utils.feature_importances import compute_surrogate_feature_importance as compute_feature_importance
    else:
        raise ValueError("Invalid strategy")
    
    
    proxy = BinaryClassificationProxy(predict=model.predict, predict_proba=model.predict_proba, classes=model.classes_)
    importances  = compute_feature_importance(proxy=proxy, X=test['X'], y=test['y'])
    ranked_importances = importances.top_alpha(0.8)
    partial_dependencies = compute_partial_dependence(test['X'], features=ranked_importances.feature_names, proxy=proxy)
    conditional_importances  = compute_feature_importance(proxy=proxy, X=test['X'], y=test['y'], conditional=True)
    return proxy, importances, ranked_importances, conditional_importances, partial_dependencies


@pytest.mark.parametrize("strategy, alpha_imp_score, xai_ease_score, position_parity, rank_alignment", [
    ("permutation", 0.010309278350515464, 1.0, 0.0, 0.0),
    ("surrogate",  0.010309278350515464, 1.0, 1.0, 1.0)
])
def test_xai_classification_metrics(strategy, alpha_imp_score, xai_ease_score, position_parity, rank_alignment, input_data):
    model, test = input_data
    proxy, importances, ranked_importances, conditional_importances, partial_dependencies = get_classification_features(model, test, strategy)

    metrics = classification_explainability_metrics(importances, partial_dependencies, conditional_importances, X=test['X'], y_pred=proxy.predict(test['X']))
    assert np.isclose(metrics.loc['Rank Alignment'].value, rank_alignment)
    assert np.isclose(metrics.loc['Position Parity'].value, position_parity)
    assert np.isclose(metrics.loc['XAI Ease Score'].value, xai_ease_score)
    assert np.isclose(metrics.loc['Alpha Importance Score'].value, alpha_imp_score)


def test_xai_classification_metrics_separated(input_data):
    model, test = input_data

    proxy, importances, ranked_importances, conditional_importances, partial_dependencies = get_classification_features(model, test, 'permutation')
        
    value = rank_alignment(conditional_importances, ranked_importances)
    assert np.isclose(value, 0.0)

    value = position_parity(conditional_importances, ranked_importances)
    assert np.isclose(value, 0.0)
    
    value = xai_ease_score(partial_dependencies, ranked_importances)
    assert np.isclose(value, 1.0)

    value = alpha_score(importances)
    assert np.isclose(value, 0.010309278350515464)


def test_rank_alignment():
    from holisticai.explainability.metrics import rank_alignment
    from holisticai.utils import ConditionalImportances, Importances

    values = {
        '0': Importances(values=[0.1, 0.2, 0.3, 0.4], 
                         feature_names=['feature_2', 'feature_3', 'feature_4']),
        '1': Importances(values=[0.4, 0.3, 0.2, 0.1], 
                         feature_names=['feature_1', 'feature_2', 'feature_3', 'feature_4'])
    }
    conditional_feature_importance = ConditionalImportances(values=values)

    ranked_feature_importance = Importances(values=[0.5, 0.3, 0.2], feature_names=['feature_1', 'feature_2', 'feature_3'])
    score = rank_alignment(conditional_feature_importance, ranked_feature_importance)
    assert np.isclose(score,0.6944444444444444)

def test_position_parity():
    from holisticai.utils import ConditionalImportances, Importances
    from holisticai.explainability.metrics import position_parity
    import numpy as np

    values = np.array([0.50, 0.40, 0.10])
    feature_names = ["feature_1", "feature_2", "feature_3"]
    feature_importance = Importances(values=values, feature_names=feature_names)

    values = {
    "group1": Importances(values=np.array([0.40, 0.35, 0.25]),
                           feature_names=["feature_1", "feature_2", "feature_3"]),
    "group2": Importances(values=np.array([0.50, 0.30, 0.20]),
                           feature_names=["feature_3", "feature_2", "feature_1"]),
    }
    conditional_feature_importance = ConditionalImportances(values=values)
    score = position_parity(conditional_feature_importance, feature_importance)
    assert np.isclose(score,0.6388888888888888)

def test_alpha_score():
    from holisticai.utils import Importances
    from holisticai.explainability.metrics import alpha_score

    values = np.array([0.10, 0.20, 0.30])
    feature_names = ["feature_1", "feature_2", "feature_3"]
    feature_importance = Importances(values=values, feature_names=feature_names)
    score = alpha_score(feature_importance)
    assert np.isclose(score, 0.6666666666666666)

def test_xai_ease_score():
    from holisticai.utils import PartialDependence, Importances
    from holisticai.explainability.metrics.global_importance import xai_ease_score
    
    partial_dependence = [[
        {
            'average': [[0.1, 0.2, 0.3, 0.1, 0.2, 0.3, 0.1, 0.2, 0.3]],
            'grid_values': [[1,2,3,4,5,6,7,8,9]]
        },
        {
            'average': [[0.4, 0.5, 0.6, 0.6, 0.5, 0.4, 0.4, 0.5, 0.6]], 
            'grid_values': [[1,2,3,4,5,6,7,8,9]]
        },
    ]]
    partial_dependence = PartialDependence(values=partial_dependence)
    feature_importance = Importances(values=np.array([0.5, 0.5]),
    feature_names=['feature1', 'feature2'])
    score = xai_ease_score(partial_dependence, feature_importance)
    assert np.isclose(score, 0.5)

def test_spread_ratio():
    from holisticai.utils import Importances
    from holisticai.explainability.metrics.global_importance import spread_ratio
    
    values = np.array([0.10, 0.20, 0.30])
    feature_names = ["feature_1", "feature_2", "feature_3"]
    feature_importance = Importances(values=values, feature_names=feature_names)
    score = spread_ratio(feature_importance)
    assert np.isclose(score, 0.9206198357143052)

def test_spread_divergence():
    from holisticai.utils import Importances
    from holisticai.explainability.metrics.global_importance import spread_divergence
    
    values = np.array([0.10, 0.20, 0.30])
    feature_names = ["feature_1", "feature_2", "feature_3"]
    feature_importance = Importances(values=values, feature_names=feature_names)
    score = spread_divergence(feature_importance)
    assert np.isclose(score, 0.8196393599933761)


from __future__ import annotations

from typing import Union

import pandas as pd
from holisticai.explainability.metrics.global_importance import (
    AlphaScore,
    PositionParity,
    RankAlignment,
    SpreadDivergence,
    SpreadRatio,
    XAIEaseScore,
)
from holisticai.explainability.metrics.global_importance._surrogate import surrogate_mean_squared_error
from holisticai.explainability.metrics.local_importance import DataStability, FeatureStability
from holisticai.utils import ConditionalImportances, Importances, LocalImportances, PartialDependence


def regression_explainability_metrics(
    importances: Importances,
    partial_dependencies: PartialDependence,
    conditional_importances: ConditionalImportances,
    X: Union[pd.DataFrame, None] = None,
    y_pred: Union[pd.Series, None] = None,
    local_importances: Union[LocalImportances, None] = None,
) -> pd.DataFrame:
    ranked_importances = importances.top_alpha(0.8)
    results = []
    metric = AlphaScore()
    value = metric(importances)
    results.append({"metric": metric.name, "value": value, "reference": metric.reference})

    metric = XAIEaseScore()
    value = metric(partial_dependencies, ranked_importances)
    results.append({"metric": metric.name, "value": value, "reference": metric.reference})

    metric = PositionParity()
    value = metric(conditional_importances, ranked_importances)
    results.append({"metric": metric.name, "value": value, "reference": metric.reference})

    metric = RankAlignment()
    value = metric(conditional_importances, ranked_importances)
    results.append({"metric": metric.name, "value": value, "reference": metric.reference})

    metric = SpreadRatio()
    value = metric(importances)
    results.append({"metric": metric.name, "value": value, "reference": metric.reference})

    metric = SpreadDivergence()
    value = metric(importances)
    results.append({"metric": metric.name, "value": value, "reference": metric.reference})

    if "surrogate" in importances.extra_attrs:
        y_surrogate = importances.extra_attrs["surrogate"].predict(X)
        value = surrogate_mean_squared_error(y_pred, y_surrogate)
        results.append({"metric": "Surrogate Accuracy Score", "value": value, "reference": 1})

    if local_importances is not None:
        metric = DataStability()
        value = metric(local_importances)
        results.append({"metric": metric.name, "value": value, "reference": metric.reference})

        metric = FeatureStability()
        value = metric(local_importances)
        results.append({"metric": metric.name, "value": value, "reference": metric.reference})
    return pd.DataFrame(results).set_index("metric")



from typing import Iterable

import matplotlib.pyplot as plt
import numpy as np
from holisticai.utils import Importances
from sklearn.tree._export import _MPLTreeExporter


def _color_brew(n):
    """Generate n colors using the 'viridis' colormap.

    Parameters
    ----------
    n : int
        The number of colors required.

    Returns
    -------
    color_list : list, length n
        List of n tuples of form (R, G, B) being the components of each color.
    """
    cmap = plt.get_cmap("viridis")
    color_list = []

    for i in range(n):
        color = cmap(0.075 + 0.875 * i / n)[:3]  # Get RGB values from cmap
        if color[0] > 1 or color[1] > 1 or color[2] > 1:
            raise ValueError("Color values must be in the range [0, 1] 1")
        if color[0] < 0 or color[1] < 0 or color[2] < 0:
            raise ValueError("Color values must be in the range [0, 1] 0")
        rgb = (int(color[0] * 255), int(color[1] * 255), int(color[2] * 255))
        color_list.append(rgb)

    return color_list


class DTExporter(_MPLTreeExporter):
    def get_fill_color(self, tree, node_id):
        # Fetch appropriate color for node
        if "rgb" not in self.colors:
            # Initialize colors and bounds if required
            self.colors["rgb"] = _color_brew(tree.n_classes[0])
            if tree.n_outputs != 1:
                # Find max and min impurities for multi-output
                self.colors["bounds"] = (np.min(-tree.impurity), np.max(-tree.impurity))
            elif tree.n_classes[0] == 1 and len(np.unique(tree.value)) != 1:
                # Find max and min values in leaf nodes for regression
                self.colors["bounds"] = (np.min(tree.value), np.max(tree.value))
        if tree.n_outputs == 1:
            node_val = tree.value[node_id][0, :]
            if tree.n_classes[0] == 1 and isinstance(node_val, Iterable) and self.colors["bounds"] is not None:
                # Unpack the float only for the regression tree case.
                # Classification tree requires an Iterable in `get_color`.
                node_val = node_val.item()
        else:
            # If multi-output color node by impurity
            node_val = -tree.impurity[node_id]
        return self.get_color(node_val)


def plot_tree(
    decision_tree,
    *,
    max_depth=None,
    feature_names=None,
    class_names=None,
    label="all",
    impurity=True,
    node_ids=False,
    precision=3,
    ax=None,
    fontsize=15,
    proportion=True,
    filled=True,
    rounded=True,
):
    exporter = DTExporter(
        max_depth=max_depth,
        feature_names=feature_names,
        class_names=class_names,
        label=label,
        filled=filled,
        impurity=impurity,
        node_ids=node_ids,
        proportion=proportion,
        rounded=rounded,
        precision=precision,
        fontsize=fontsize,
    )
    return exporter.export(decision_tree, ax=ax)


def plot_surrogate(feature_importance: Importances, ax=None, **kargs):
    """
    Plots the surrogate tree for feature importance.

    Parameters
    ----------
    feature_importance: Importances
        The feature importance object.
    ax: (matplotlib.axes.Axes, optional)
        The matplotlib axes to plot the tree on. If not provided, a new figure and axes will be created.
    kargs:
        Additional keyword arguments to be passed to the `plot_tree` function.

    Returns
    -------
    ax: matplotlib.axes.Axes

    Example
    -------
    >>> plot_surrogate(feature_importance)

    The plot should look like this:

    .. image:: /_static/images/xai_plot_surrogate.png
        :alt: Plot Surrogate

    """

    if "surrogate" not in feature_importance.extra_attrs:
        raise ValueError("Surrogate key does not exist in feature_importance.extra_attrs")

    if ax is None:
        _, ax = plt.subplots(1, 1, figsize=(30, 10))

    plot_tree(
        feature_importance.extra_attrs["surrogate"],
        feature_names=feature_importance.values,
        max_depth=3,
        ax=ax,
        **kargs,
    )
    description = """Classification: Color indicate majority class.\nRegression: Color indicate extremity of values."""
    ax.text(0.02, 0.92, description, fontsize=15, ha="left", transform=plt.gca().transAxes)
    return ax



from __future__ import annotations

from typing import Optional, Union

import numpy as np
import pandas as pd


def x_array_to_df(x_arr, feature_names):
    return pd.DataFrame(x_arr, columns=feature_names)


def x_to_nd_array(x: pd.DataFrame):
    return np.array(x)


def to_categorical(labels: Union[np.ndarray, list[float]], nb_classes: Optional[int] = None) -> np.ndarray:
    """
    Convert an array of labels to binary class matrix.

    Parameters
    ----------
    labels : Union[np.ndarray, list[float]]
        An array of integer labels of shape `(nb_samples,)`.
    nb_classes : int, optional
        The number of classes (possible labels), by default None.

    Returns
    -------
    np.ndarray
        A binary matrix representation of `y` in the shape `(nb_samples, nb_classes)`.
    """
    labels = np.array(labels, dtype=int)
    if nb_classes is None:
        nb_classes = np.max(labels) + 1
    categorical = np.zeros((labels.shape[0], nb_classes), dtype=np.float32)
    categorical[np.arange(labels.shape[0]), np.squeeze(labels)] = 1
    return categorical


def format_function_predict_proba(learning_task, predict_proba_fn):
    """
    Format the predict_proba function based on the learning task.

    Parameters
    ----------
    learning_task : str
        The learning task.
    predict_proba_fn : callable
        The predict_proba function.

    Returns
    -------
    callable
        The formatted predict_proba function.
    """
    if learning_task == "binary_classification":

        def forward(x: np.ndarray, feature_names: list[str]):
            x_df = x_array_to_df(x, feature_names=feature_names)
            score = np.array(predict_proba_fn(x_df))
            if score.ndim == 2:
                return score
            return np.stack([1 - score, score], axis=1)

    return forward



from holisticai.datasets import load_dataset

def test_load_dataset():
    SHARD_SIZE = 50
    dataset = load_dataset("adult", protected_attribute='sex').sample(1000)
    dataset = dataset.map(lambda x: {'new': str(x['group_a'])  + str(x['y'])}, vectorized=False)
    dataset = dataset.groupby(['y','group_a']).head(SHARD_SIZE)
    dataset = dataset.train_test_split(test_size=0.2, stratify=dataset['y'], random_state=0)


def test_map():
    dataset = load_dataset('us_crime')
    dataset = dataset.train_test_split(test_size=0.2, random_state=42)
    test = dataset['test']
    new_test = test.map(lambda sample: {'group': sample['y']}, vectorized=False)
    assert list(new_test['X'].columns) == list(test['X'].columns)


from __future__ import annotations

import numpy as np
import pandas as pd


def generate_rankings(M: int | None, k: int | None, p: float | None, return_p_attr=False, seed: int | None = 42):
    """
    Generates M rankings of k elements (candidates) using Yang-Stoyanovich process

    Parameters
    ----------
    M: int
        Number of rankings to generate
    k: int
        Number of elements should each ranking have
    p: float
        Probability that a candidate is protected
    return_p_attr: bool
        Whether to return protected attribute
    seed: int
        Random seed

    Returns
    -------
    DataFrame, Tuple
        DataFrame or Tuple of DataFrames
    """
    random_state = np.random.RandomState(seed)
    rankings = []
    for m in range(M):
        for i in range(k):
            is_protected = random_state.random() <= p
            rankings.append({"X": m, "Y": k - i, "score": k - i, "protected": is_protected})

    df = pd.DataFrame(rankings)
    if return_p_attr:
        return df[["X", "Y", "score"]], df[["X", "Y", "protected"]]
    return df[["X", "Y", "score", "protected"]]
